from . import eNMR_Methods

class Simulated(eNMR_Methods):
    '''
    sublass of eNMR_Methods and Measurement for the import of simulated data via the SpecSim class in the Phasefitting Module
    '''

    def __init__(self, sim, alias=None):
        self.g = sim.params.spec_par['g']
        self.d = sim.params.spec_par['d']
        self.Delta = sim.params.spec_par['Delta']
        self.delta = sim.params.spec_par['delta']
        self.dependency = 'U'
        self.eNMRraw = sim.eNMRraw
        self.data = sim.data
        self.data_orig = sim.data
        self.ppm = sim.ppm
        self.eNMRraw['g in T/m'] = self.g

        # the gamma_values in rad/Ts
        gamma_values = {'1H':26.7513e7,
                        '7Li': 10.3962e7,
                        '19F': 25.1662e7}
        self.gamma = gamma_values[sim.params.spec_par['NUC']]
        # Umrechnung von rad in °
        self.gamma = self.gamma/2/np.pi*360
        
        self.lin_res_dic = {}
        self.alias = alias
        self.path = 'simulated data/'
        self.expno = '0'
        self.uInk = self.eNMRraw.loc[1, 'U / [V]'] - self.eNMRraw.loc[0, 'U / [V]']
        self._x_axis = 'U / [V]'
        #self._x_axis = {"U": "U / [V]",
            #"G": "g in T/m"
            #}[self.dependency.upper()]
##################################
 



"""
The package eNMRpy is a tool for the import and analysis of electrophoretic NMR spectroscopy raw data.
The package is mainly based on nmrglue and lmfit other common python imports.

A peer reviewed paper showing the performance and advantages of this package is available under:
Schmidt, Florian, et al. "Spectral deconvolution in electrophoretic NMR to investigate the migration of
neutral molecules in electrolytes." Magnetic Resonance in Chemistry 58.3 (2020): 271-279.

The import classes are grouped in the submodule eNMRpy.Measurement

While the phase analysis via zero order phase correction is included in the import classes,
phase analysis via lorentzian fits is available under eNMRpy.Phasefitting, and via 2D Fast Fourier Transformation
under eNMRpy.MOSY
"""

#from .Phasefitting import SpecModel
#from .MOSY import MOSY
# Pavel Class as the Std Import-Class since it is mostly used in our working group
from .Measurement import Pavel as Import_eNMR_Measurement 
from . import tools
#print('%s imported'%__name__)




## coding: utf-8
# This is the eNMR class. THE library for the evaluation of bruker-eNMR-spectra based on the VC-PowerSource of the
# Schönhoff working group.
# It works with the volt-increment method which calculates the respective voltage with the VC-list
# Further Implementation can be asked for at f_schm52@wwu.de
from .eNMR_Methods import _eNMR_Methods
import matplotlib.pyplot as plt
from .base import Measurement
from re import findall
import numpy as np
import pandas as pd
     

#class eNMR_Emma(eNMR_Measurement):
class eNMR_Emma(_eNMR_Methods):
    #'''
    #This is the subsubclass of Masurement() and subclass of eNMR_Methods specialised to process data obtained from the experimental Schönhoff set-up
    
    #path:
        #relative or absolute path to the measurements folder
    #measurement:
        #the to the experiment corresponding EXPNO
    #alias:
        #Here you can place an individual name relevant for plotting. If None, the path is taken instead.    
    #Uink:
        #voltage increment. Usually extracted from the title file if defined with e.g. "Uink = 10V"
        #If Uink cannot be found or is wrong it can be entered manually during the data import.
        #The voltage list is calculated from the voltage increment and the vc list when the incrementation loop is used in the pulse program
    #dependency:
        #'U': voltage dependent eNMR measurement
        #'G': fieldgradient dependent eNMR measurement

    #linebroadening:
        #setting a standard-value for the linebroadening.
    #'''
    def __init__(self, path, expno, Uink=None, dependency="U", alias=None, lineb=0.5, electrode_distance=2.2e-2):
        Measurement.__init__(self, path, expno, lineb=lineb, alias=alias)
        self.dependency = dependency.upper()
        
        self._x_axis = {"U": "U / [V]",
                   "G": "g in T/m",
                   "I": "I / mA",
                   'RI': 'RI / V'
                   }[self.dependency.upper()]
        
        #self._x_axis = {"G": "g in T/m",
                        #"U": "U / [V]"}[self.dependency.upper()]
        
        self.difflist = pd.read_csv(self.dateipfad+"/difflist",
                            names=["g in T/m"])*0.01
        
        self.vcList = pd.DataFrame()
        
        if self.dic['acqus']['PULPROG'][-3:] == 'var':
            polarity = 1
            print('this is a regular measurement! (non-_pol)')
        elif self.dic['acqus']['PULPROG'][-3:] == 'pol':
            polarity = -1
            print('this is a _pol-Measurement!')
        else:
            print("no var or pol PULPROG")
    
        if dependency.upper() == "U":
            try:
                # takes the title page to extract the volt increment
                title = open(self.dateipfad+"/pdata/1/title").read()
                # gets the voltage increment using a regular expression
                #uimport = findall('[U|u]in[k|c]\s*=?\s*\d+', title)[0]
                uimport = findall('[U|u]in[k|c]\s*=+\s*\d+', title)[0]
                self.uInk = int(findall('\d+', uimport)[0])
            except ValueError:
                print('no volt increment found\nyou may want to put it in manually')
                self.uInk = Uink
            except IndexError:
                print('No Uink found! May not be an eNMR experiment.')
                self.uInk = Uink
            
            self.vcList["U / [V]"] = [i*self.uInk*polarity for i in range(len(self.difflist))]
                
        elif dependency.upper() == "G":
            try:
                # takes the title page to extract the volt increment
                title = open(self.dateipfad+"/pdata/1/title").read()
                # gets the voltage increment using a regular expression
                uimport = findall('[U|u]\s*=?\s*\d+', title)[0]
                self.uInk = int(findall('\d+', uimport)[0])

            except ValueError:
                print('no volt increment found\nyou may want to put it in manually')
                self.uInk = Uink
            except IndexError:
                print('No Uink found! May not be an eNMR experiment.')
                self.uInk = Uink # Uinktext
                
            #if Uink is not None:
                #self.uInk = Uink
            
            self.vcList["U / [V]"] = [self.uInk*polarity for i in range(len(self.difflist))]
            #self.vcList["U / [V]"] = [self.vcList["vc"][n]/2*self.uInk if self.vcList["vc"][n] % 2 == 0
                                    #else (self.vcList["vc"][n]+1)/2*self.uInk*-1
                                    #for n in range(len(self.data[:, 0]))]

        #if self.dependency.upper() == "U":
            #try:
                #self.vcList = pd.read_csv(self.dateipfad+"/vclist",
                                          #names=["vc"]).loc[:len(self.data[:, 0])-1]
                
            #except:
                #print("There is a Problem with the VC-list or you performed a gradient dependent measurement")
        #elif self.dependency.upper() == "G":
            #self.vcList = pd.DataFrame(np.ones((len(self.data[:, 0]), 1)),
                                       #columns=["vc"])
        #else:
            #print("The dependency is not properly selected, try again!")

        #self.difflist = pd.read_csv(self.dateipfad+"/difflist",
                                    #names=["g in T/m"])*0.01
        
            #if Uink is not None:
                #self.uInk = Uink
                
            #self.vcList["U / [V]"] = [self.vcList["vc"][n]/2*self.uInk if self.vcList["vc"][n] % 2 == 0
                                    #else (self.vcList["vc"][n]+1)/2*self.uInk*-1
                                    #for n in range(len(self.data[:, 0]))]
            
        # try to open phase data, otherwise create new
        try:
            self.eNMRraw = pd.read_csv(self.path+"phase_data_"+self.expno+".csv",
                                       index_col=0, sep=" ")
            # --> update voltage list
            self.eNMRraw["U / [V]"] = self.vcList["U / [V]"]
        except:
            print("eNMRraw was missing and is generated")
            self.vcList["ph0"] = np.zeros(len(self.data.real[:, 0]))
            self.eNMRraw = self.vcList
        finally:
            self.eNMRraw["g in T/m"] = self.difflist
        
        self.p1 = self.dic["acqus"]["P"][1]
        self.d1 = self.dic["acqus"]["D"][1]
        
        try:
            # import of diffusion parameters for newer Spectrometers
            import xml.etree.ElementTree as etree
            diffpar = etree.parse(self.dateipfad+'/diff.xml')
            root = diffpar.getroot()
            self.Delta = float(root.findall('DELTA')[0].text)*1e-3
            self.delta = float(root.findall('delta')[0].text)*1e-3  # it should be read as in microseconds at this point due to bruker syntax
            print('The diffusion parameters were read from the respectie .XML!')
        except:
            # determination of the diffusion parameters for Emma
            self._d2 = self.dic["acqus"]["D"][2]
            self._d5 = self.dic["acqus"]["D"][5]
            self._d9 = self.dic["acqus"]["D"][9]
            self._d11 = self.dic["acqus"]["D"][11]
            self._p19, self._p18, self._p17 = self.dic["acqus"]["P"][19],\
                                            self.dic["acqus"]["P"][18],\
                                            self.dic["acqus"]["P"][17]
            print('That did not work. Your data is from an old spectrometer!')
            # calculating usable parameters
            self.delta = self._p17+self._p18
            self._Delta_1 = 0.001*(self._p17*2+self._p18)+(self._d2+self._d9+self._d5+self._d11)*1000+0.001*self.p1+self._d11
            self._Delta_2 = 0.001*(self._p17*2+self._p18)+(self._d2+self._d9+self._d5+self._d11)*1000+0.001*self.p1*2
            self._spoiler = (self._d11+self._p17+self._p19+self._p17)*0.001+self._d2*1000
            self.Delta = self._Delta_1+self._Delta_2+2*self._spoiler
            self.Delta *=1e-3
            self.delta *=1e-6
            

        # Elektrodenabstand in m
        self.d = electrode_distance
        self.g = self.eNMRraw["g in T/m"][0]
    
    def __add__(self, other):
        
        for obj in [self, other]:
            for k in obj.eNMRraw.columns:
                if k[:2] == 'ph':
                    obj.eNMRraw[k] -= obj.eNMRraw.loc[0, k]
                    print('%s normalized to 0V'%k)
                else:
                    pass
        self.eNMRraw = self.eNMRraw.append(other.eNMRraw)
        self.eNMRraw.sort_values('U / [V]', inplace=True)
        return self



import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

class MOSY(object):
    """
    takes an eNMR Measurement obj and yields as MOSY object for MOSY processing and depiction.
    the Measurement object needs to be processed (fouriertransformed in F2) before passing it to this class
    """
    def __init__(self, obj):
        self.obj = obj
        self.eNMRraw = obj.eNMRraw
        self.eNMRraw['data'] = None
        self.eNMRraw['fid'] = None
        for row in range(len(obj.data[:,0])):
            self.eNMRraw.at[row, 'data'] = pd.Series(obj.data[row,:]) # new syntax
            #self.eNMRraw.set_value(row, 'data', pd.Series(obj.data[row,:])) # old syntax
            #self.eNMRraw.set_value(row, 'fid', pd.Series(obj.fid[row,:]))
        data_sorted = np.array(obj.eNMRraw['data'].tolist())
        self.data = data_sorted.astype('complex').copy()
        self.TD1 = len(self.data[:,0])
        self.mscale = None
        self.mX = None
        self.mY = None
#         self.mZ = None
        
    def zerofilling_old(self, n=128, dimension=0):
        """
        n: number of total datapoints along the F1-dimension (Voltage)
        dimension 0 = F1
        dimension 1 = F2
        """
        if dimension == 0:
            self.data = np.concatenate((self.data, 
                                [[0 for x in range(len(self.data[0,:]))] for y in range(n-len(self.data[:,0]))]))
        if dimension == 1:
            self.data = np.concatenate((self.data, 
                                [[0 for x in range(n-len(self.data[0,:]))] for y in range(len(self.data[:,0]))]), axis=1)
        print('got it!')
        
    def zerofilling(self, n=128, dimension=0):
        """
        n: number of total datapoints along the F1-dimension (Voltage)
        dimension 0 = F1
        dimension 1 = F2
        """
        print('zero filling started!')
        
        if dimension == 0:
            old_shape = self.data.shape
            new_shape = old_shape
            new_shape = n, old_shape[1]
            print(old_shape, new_shape)
            data_new = np.zeros(new_shape, dtype=complex)
            for i in range(old_shape[0]):
                data_new[i,:] = self.data[i,:]
            self.data = data_new
            del(data_new)
            
        if dimension == 1:
        
            old_shape = self.data.shape
            new_shape = old_shape
            new_shape = old_shape[0], n
            print(old_shape, new_shape)
            data_new = np.zeros(new_shape, dtype=complex)
            for i in range(old_shape[1]):
                data_new[:,i] = self.data[:,i]
            self.data = data_new
            del(data_new)
        
        print('zero filling finished!')

    def fft_F1 (self):
        """
        Fourier Transformation from nmrglue.proc_base.fft)
        along the F1 dimension
        """
        
        from nmrglue import proc_base
        
        #data_temp = np.zeros(self.data.shape)
        for n in range(len(self.data[0,:])):
            self.data[:,n] = proc_base.fft(self.data[:,n])
        print("done")
    
    def calc_MOSY(self, u_max = None, n_zf=2**12, mobility_scale=True, include_0V=True, electrode_distance=2.2e-2, old_zf=False):
        
        '''
        Calculates the MOSY according to the States-Haberkorn-Ruben method described for MOSY applications by AUTOR NENNEN!
        
        u_max:
            maximum Voltage to be used for processing.
        n_zf:
            number of zerofilling for the F1 dimension
        mobility_scale:
            prints the converted mobilty y-scale if True
        include_0V:
            include the 0V spectrum or not.
        '''
        x = self.obj._x_axis
        
        if u_max is None:
            u_max = 10000
        
        if include_0V:
            positive = self.eNMRraw[(self.eNMRraw[x] >= 0)
                                    &(self.eNMRraw[x] <= u_max)].sort_values(x)
            negative = self.eNMRraw[(self.eNMRraw[x] <= 0)
                                    &(self.eNMRraw[x] >= -u_max)].sort_values(x, ascending=False)
        else:
            positive = self.eNMRraw[self.eNMRraw[x] >0].sort_values(x)
            negative = self.eNMRraw[self.eNMRraw[x] <0].sort_values(x, ascending=False)

        
        #dataset conversion for the SHR-method
        #if include_0V:
            #positive = self.eNMRraw[(self.eNMRraw['U / [V]'] >= 0)
                                    #&(self.eNMRraw['U / [V]'] <= u_max)].sort_values('U / [V]')
            #negative = self.eNMRraw[(self.eNMRraw['U / [V]'] <= 0)
                                    #&(self.eNMRraw['U / [V]'] >= -u_max)].sort_values('U / [V]', ascending=False)
        #else:
            #positive = self.eNMRraw[self.eNMRraw['U / [V]'] >0].sort_values('U / [V]')
            #negative = self.eNMRraw[self.eNMRraw['U / [V]'] <0].sort_values('U / [V]', ascending=False)

        SHR_real = np.zeros((len(positive['data']), len(positive['data'].iloc[0])))
        SHR_imag = np.zeros((len(positive['data']), len(positive['data'].iloc[0])))

        for n in range(1, len(positive['data'])):
            SHR_real[n,:] = positive['data'].iloc[n].values.real + negative['data'].iloc[n].values.real
            SHR_imag[n,:] = positive['data'].iloc[n].values.imag - negative['data'].iloc[n].values.imag

        SHR = SHR_real + SHR_imag*1j
        del(SHR_real)
        del(SHR_imag)
        self.data = SHR
        del(SHR)
        if old_zf:
            self.zerofilling_old(n=n_zf)
        else:
            self.zerofilling(n=n_zf)
        self.fft_F1()
     
        X = np.array([self.obj.ppm for y in range(len(self.data[:,0]))])
        Y = np.array([[y for x in range(len(self.data[0,:]))] for y in range(len(self.data[:,0]))])
        
        Y = ((0.5*n_zf-Y)*self.TD1/n_zf*360)/self.TD1 # conversion to ° phasechange per Uink
        Y = Y/self.obj.uInk  # ° phasechange per Volt
        
        if mobility_scale:
            #Y = (Y*self.obj.d)/(self.obj.gamma*self.obj.delta*self.obj.Delta*self.obj.g) # old version with self.d
            Y = (Y*electrode_distance)/(self.obj.gamma*self.obj.delta*self.obj.Delta*self.obj.g)
        
        self.mX = X
        del(X)
        self.mY = -Y
        del(Y)
    
    
    def plot_MOSY(self, xlim=None, ylim=(-1e-8, 1e-8), yscale='linear',
                  save=False, savepath='',
                  tight_layout=False, dpi=300, y_autoscale=True,
                  h_ratio=7, w_ratio=8, figsize=(5,5),
                  latex_backend=False, hspace=0, wspace=0,
                  **kwargs):
        
        from matplotlib import gridspec
        """
        yscale:
            .. ACCEPTS: [ 'linear' | 'log' | 'symlog' | 'logit' | ... ]
        **kwargs:
            are passed to the .contour()
            for example:
                levels: list of values to be drawn as height lines
                
        :returns: figure
        """
        fig= plt.figure(figsize=figsize)
        
        gs = gridspec.GridSpec(2, 2, width_ratios=[1, w_ratio], height_ratios=[1,h_ratio]) 

        mitte = fig.add_subplot(gs[3])
        oben = fig.add_subplot(gs[1], sharex=mitte, frameon=False)
        links = fig.add_subplot(gs[2], sharey=mitte, frameon=False)

        #fig.subplots_adjust(hspace=0, wspace=0)

        mitte.yaxis.tick_right()
        mitte.contour(self.mX, self.mY, self.data.real, **kwargs)
        mitte.set_ylim(ylim)
        mitte.set_xlim(xlim)
        links.set_xticks([])

        # calculation and plotting of the projections taking the maxima
        links.plot(np.amax(self.data, axis=1), self.mY[:,0],  'k')
        links.set_xlim(links.get_xlim()[::-1])
        oben.plot(self.mX[0], np.amax(self.data, axis=0), 'k')
        
        #axis format
        oben.tick_params(bottom='off')
        links.tick_params(left='off')
        oben.set_yticks([])
        plt.setp(oben.get_xticklabels(), visible=False);
        plt.setp(links.get_yticklabels(), visible=False);
        plt.setp(links.get_yaxis(), visible=False);
        if latex_backend:
            mitte.set_ylabel(r'$\mu\;/\;(\textrm{m}^2 \textrm{V}^{-1} \textrm{s}^{-1})$')
        else:
            mitte.set_ylabel(r'$\mu$ / (m$^2$V$^{-1}$s$^{-1})$')
        mitte.yaxis.set_label_position("right")
        mitte.set_xlabel(r'$\delta$ / ppm')
        mitte.set_yscale(yscale)
        mitte.ticklabel_format(axis='y', style='sci', scilimits=(0,2))
        mitte.grid()
        
        fig.subplots_adjust(hspace=hspace, wspace=wspace)
        
        def autoscale_y(bx, margin=0.1):
            """This function rescales the y-axis based on the data that is visible given the current xlim of the axis.
            ax -- a matplotlib axes object
            margin -- the fraction of the total height of the y-data to pad the upper and lower ylims"""

            def get_bottom_top(line):
                xd = line.get_xdata()
                yd = line.get_ydata()
                lo,hi = bx.get_xlim()
                try:
                    y_displayed = yd[((xd>lo) & (xd<hi))]
                    h = np.max(y_displayed) - np.min(y_displayed)
                except:
                    y_displayed = yd[((xd<lo) & (xd>hi))]
                    h = np.max(y_displayed) - np.min(y_displayed)
                
                bot = np.min(y_displayed)#-margin*h
                top = np.max(y_displayed)#+margin*h
                print(bot, top, h)
                return bot,top

            lines = bx.get_lines()
            bot,top = np.inf, -np.inf

            for line in lines:
                new_bot, new_top = get_bottom_top(line)
                if new_bot < bot: bot = new_bot
                if new_top > top: top = new_top

            bx.set_ylim(bot,top)
            
        if y_autoscale:
            autoscale_y(oben, xlim)
        
        if tight_layout:
            fig.tight_layout()
        
        if save:
            fig.savefig(savepath, dpi=dpi)

        return fig
    
    def plot_slices_F1(self, ppm, xlim=None, scaling=None, normalize=False, vline_0=False, annotate=False, annotate_pos=None, legend_loc=None, latex_backend=False, colors=None, figsize=None):
        """
        plots slices in F1-direction

        ppm:
            list of chemical shift-slices to be plotted
        xmin, xmax:
            limits for the x-axis
        scaling:
            list of scaling factors for spectra to adjust intensities
        normalize:
            normalization of each peak to 1
        vline_0:
            prints a vertical line at µ = 0
        annotate:
            automatically annotates the maxima of the slices and displays the respective mobility
        colors:
            list of colors/colorcodes as strings
        
        :returns: figure
        """

        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
        # getting the standard color cycle
        prop_cycle = plt.rcParams['axes.prop_cycle']
        if colors is None:
            colors = prop_cycle.by_key()['color']

        if scaling is None:
            scaling = np.ones((len(ppm)))
        i=0
        # setting the initial textposition depending on the number of slices
        xpostext = 0.5-((len(ppm)-1)*0.1-0.1)
        if xpostext < 0.2:
            xpostext = 0.2
        ypostext = 0.2
        
        for n in ppm:
            pos = np.searchsorted(self.obj.ppm, n)
            x = self.mY[:,pos]
            if normalize:
                y = scaling[i]*self.data[:,pos].real/self.data[:,pos].max()
                ax.plot(x, y, label=r'$\delta=%.2f$'%n, c=colors[i])
                ax.set_ylabel('normalized intensity')
            else:
                y = scaling[i]*self.data[:,pos].real
                ax.plot(x, y, label=r'$\delta=%.2f$'%n, c=colors[i])
                ax.set_ylabel('intensity / a.u.')
            
            if annotate:
                xmax = x[np.argmax(y)]
                ymax = y.max()
                text = "$\mu$=%.2E"%(xmax)
                bbox_props = dict(boxstyle="square,pad=0.3", fc="w", ec="k", lw=0.72)
                arrowprops=dict(arrowstyle="->", color=colors[i])#, connectionstyle="angle,angleA=0,angleB=60")
                kw = dict(xycoords='data',textcoords="axes fraction",
                arrowprops=arrowprops, bbox=bbox_props, ha="right", va="top")
                if type(annotate_pos) == list:
                    xpostext, ypostext = annotate_pos[i]
                elif type(annotate_pos) == tuple:
                    xpostext += annotate[0]
                    ypostext += annotate[1]
                ax.annotate(text, xy=(xmax, ymax), xytext=(xpostext,ypostext), **kw)
                #placing of the textboxes
                if xpostext >= 1:
                    xpostext = 0.2
                    ypostext += 0.1
                else:
                    xpostext += 0.2
            i += 1

        if vline_0:
            ax.vlines(0, *ax.get_ylim(), linestyles='dotted')
        if latex_backend:
            ax.set_xlabel('$\mu$\;/\;(m$^2$V$^{-1}$s$^{-1})$')
        else:
            ax.set_xlabel('$\mu$ / (m$^2$V$^{-1}$s$^{-1})$')
        ax.set_xlim(xlim)
        ax.legend(loc=legend_loc)
        ax.ticklabel_format(style='sci', scilimits=(0,3))
        return fig
    
    def export_data(self, path):
        """
        saves the X, Y and Z-data to 3 textfiles in order to be used with other programs
        """
        pX = path+'_X'
        pY = path+'_Y'
        pZ = path+'_Z'
        np.savetxt(pX, self.mX, delimiter=',')
        np.savetxt(pY, self.mY, delimiter=',')
        np.savetxt(pZ, self.data, delimiter=',')
 



import matplotlib.pyplot as plt
import nmrglue as ng
from re import findall
import numpy as np
import pandas as pd
from io import StringIO

class Measurement(object):
    """
    This is the base class for the analysis of eNMR raw spectra obtained on Bruker spectrometers with given settings.
    if creating an object from an eNMR measurement creates an error, Uink might have failed to be set and needs to be set manually instead.
    
    path:
        relative or absolute path to the measurements folder
    measurement:
        the to the experiment corresponding EXPNO
    alias:
        Here you can place an individual name relevant for plotting. If None, the path is taken instead.
        
    lineb:
        linebroadening
    """

    def __init__ (self, path, expno, alias=None,  lineb=.3, n_zf_F2=2**14):
        self.path = path
        self.expno = str(expno)
        self.dateipfad = self.path+self.expno
        self.alias = alias
        self.name = self.dateipfad.split(sep='/')[-2]+'/'+self.expno
        self.lineb = lineb
        
        self.comment_string = None
        
        # takes the title page to extract the volt increment
        try:
            title = open(self.dateipfad+"/pdata/1/title").read()
            # reformatting the title_page removing excess of newlines (\n)

        except UnicodeDecodeError:
            title = open(self.dateipfad+"/pdata/1/title", encoding='ISO-8859-15').read()
            
        title = findall('.+', title)              
        self.title_page = self.name+'\n'

        for n in title:
            self.title_page += n+'\n'        

        # read in the bruker formatted data
        self.dic, self.data = ng.bruker.read(self.dateipfad)
        self.pdata = ng.bruker.read_procs_file(self.dateipfad+"/pdata/1")
        # original dataset
        self.data_orig = self.data

        # Berechnung der limits der ppm-skala
        self._ppm_r = self.pdata["procs"]["OFFSET"]
        self._ppm_l = -(self.dic["acqus"]["SW"]-self.pdata["procs"]["OFFSET"])
        self._ppmscale = np.linspace(self._ppm_l, self._ppm_r, n_zf_F2)  # np.size(self.data[0,:]))
        self.ppm = self._ppmscale

        # Bestimmung der dictionaries für die NMR-Daten aus den Messdateien
        # --> wichtig für die Entfernung des digitalen Filters
        #self.udic = ng.bruker.guess_udic(self.dic, self.data)
        #self._uc = ng.fileiobase.uc_from_udic(self.udic)

        # getting some other variables
        self.TD = self.dic["acqus"]["TD"]
        self.fid = self.data.copy()
        self.n_zf_F2 = n_zf_F2

        # the gamma_values in rad/Ts
        gamma_values = {'1H':26.7513e7,
                        '7Li': 10.3962e7,
                        '19F': 25.1662e7}

        self.gamma = gamma_values[self.dic["acqus"]["NUC1"]]
        # conversion from rad in °
        self.gamma = self.gamma/2/np.pi*360
        
        self.nuc = self.dic["acqus"]["NUC1"]
        
        # initialize dictionary for linear regression results
        self.lin_res_dic = {}
        #self.dependency = None

        # END __INIT__
    
    def ph_ref(self, ref):
        '''
        reference your phase shifts on the shifts of one peak.
        ref: string, name of the peak which phase is used for reference (usually solvent).
        '''

        ref_orig_ph = self.eNMRraw[ref].copy
        for i in filter(lambda x: x[:2] =='ph' and x[-1]!='r', self.eNMRraw.keys()):
            self.eNMRraw[i] = self.eNMRraw[i] - ref_orig_ph
    
    
    def comment(self, s, append_to_title=False):
        '''
        write a comment as a multi-line-string or normal string
        '''
        self.comment_string = s
        
        if append_to_title:
            #self._title_page_orig = self.title_page
            self.title_page += self.comment_string

    def calibrate_ppm(self, ppmshift):
        """
        calibrate ppm-scale by adding the desired value.

        :param ppmshift: value added to ppm-scale
        :return: nothing
        """
        self.ppm = self._ppmscale + ppmshift

    def proc(self, linebroadening=None, phc0=0, phc1=0, zfpoints=None, xmin=None, xmax=None, cropmode="percent"):
        """
        processes the spectral data by:
            - removing digital filter
            - create separate fid data
            - linebroadening on spectral data
            - zero filling
            - fourier transformation
            
        crops the data after fft set to xmin and xmax on the x-axis and returns the value
        when xmin and xmax or not both None
        
        xmin, xmax:
            min and maximum x-values to crop the data for processing (and display)
            can take relative or absolute values depending on the cropmode.
        
        cropmode: changes the x-scale unit
            "percent": value from 0% to 100% of the respective x-axis-length --> does not fail
            "absolute": takes the absolute values --> may fail            
        
        :return: nothing
        """
        
        
        if linebroadening is None:
            linebroadening = self.lineb
        
        if zfpoints is not None:
            zfp = zfpoints
        else:
            zfp = self.n_zf_F2
        _lineb = linebroadening
        
        # remove the digital filter
        self.data = ng.bruker.remove_digital_filter(self.dic, self.data)

        # process the spectrum
        # linebroadening
        self.data = ng.proc_base.em(self.data, lb=_lineb/self.dic["acqus"]["SW_h"])

        # zero fill to 32768 points
        try:
            self.data = ng.proc_base.zf_size(self.data, zfp)
        except ValueError:
            zfp = 2**15
            self.data = ng.proc_base.zf_size(self.data, zfp)
        # Fourier transform
        self.data = ng.proc_base.fft(self.data)
        
        # Phasecorrection
        self.data = ng.proc_autophase.ps(self.data, phc0, phc1)
        #correct ppm_scale
        self._ppmscale = np.linspace(self._ppm_l, self._ppm_r, zfp)  # np.size(self.data[0,:]))
        self.ppm = self._ppmscale

        self.data_orig = self.data

        if (xmin is not None) or (xmax is not None):
            self.data = self.set_spectral_region(xmin=xmin, xmax=xmax, mode=cropmode)
    

    def plot_fid(self, xmax=None, xmin=0, step=1):
        """
        plots every n-th(step) fid and scales the time axis with xmax and xmin
        
        :returns: figure
        """
        
        _xmax = self.TD/2 if xmax is None else xmax
        _xmin = xmin

        fig, ax = plt.subplots()
        
        for n in range(len(self.data[::step, 0])):
            ax.plot(self.fid[n, ::1].real)

        ax.set_xlim((_xmin, _xmax))
        ax.set_xlabel("data points")
        ax.ticklabel_format(style='sci', scilimits=(0,0), axis='y')
        ax.legend([x*step for x in range(len(self.data[::step, 0]))],
                   ncol=3,
                   title="row",
                   loc=1)
        return fig

    def plot_spec(self, row, xlim=None, figsize=None, invert_xaxis=True, sharey=True):#, ppm=True):
        """
        plots row 0 and row n in the range of xmax and xmin
        
        :returns: figure
        """
        
        _max = None if xlim is None else xlim[0]
        _min = None if xlim is None else xlim[1]
        
        if type(xlim) != list:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
        elif type(xlim) == list:
            fig, ax = plt.subplots(ncols=len(xlim), nrows=1, figsize=figsize, sharey=sharey)
        
        _min = self.ppm[0] if xlim is None else xlim[1]
        _max = self.ppm[-1] if xlim is None else xlim[0]
        
        if type(xlim) != list:
            if type(row) == list:
                for r in row:
                    ax.plot(self.ppm, self.data[r, ::1].real, label='row %i'%r)
            else:
                ax.plot(self.ppm, self.data[row, ::1].real, label='row %i'%row)
            ax.legend()
            ax.set_xlim(xlim)
            #ax.set_title("this is row %i at %.0f V" % (row, self.eNMRraw.loc[row, "U / [V]"]))
            ax.set_xlabel("$\delta$ / ppm")
            ax.set_ylabel("intensity / a.u.")

            ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            if invert_xaxis:
                xlimits = ax.get_xlim()
                ax.set_xlim(xlimits[::-1])
        
        elif type(xlim) == list:
            for axis, xlim in zip(ax,xlim):
                if type(row) ==list:
                    for r in row:
                        axis.plot(self.ppm, self.data[r, ::1].real, label='row %i'%r)
                else:
                    axis.plot(self.ppm, self.data[row, ::1].real, label='row %i'%row)
                
                axis.set_xlim(xlim)
                #ax.set_title("this is row %i at %.0f V" % (row, self.eNMRraw.loc[row, "U / [V]"]))
                axis.set_xlabel("$\delta$ / ppm")
                axis.set_ylabel("intensity / a.u.")

                axis.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
                if invert_xaxis:
                    xlimits = axis.get_xlim()
                    axis.set_xlim(xlimits[::-1])
            #fig.legend()
        return fig
    
    def plot_spec_1d(self, xlim=None, figsize=None, invert_xaxis=True):#, ppm=True):
        """
        plots row 0 and row n in the range of xmax and xmin
        
        :returns: figure
        """
        _max = None if xlim is None else xlim[0]
        _min = None if xlim is None else xlim[1]

        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111)
        
        _min = self.ppm[0] if xlim is None else xlim[1]
        _max = self.ppm[-1] if xlim is None else xlim[0]
        
        ax.plot(self.ppm, self.data[::1].real)
        
        #ax.legend()
        ax.set_xlim(xlim)
        #ax.set_title("this is row %i at %.0f V" % (row, self.eNMRraw.loc[row, "U / [V]"]))
        ax.set_xlabel("$\delta$ / ppm")
        ax.set_ylabel("intensity / a.u.")

        ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
        if invert_xaxis:
            xlimits = ax.get_xlim()
            ax.set_xlim(xlimits[::-1])
        return fig

    def set_spectral_region(self, xmin, xmax, mode='absolute', ppm=True, original_data=True):
        """
        crops the data of the object set to xmin and xmax on the x-axis and returns the value
        
        mode: changes the x-scale unit
            "percent": value from 0% to 100% of the respective x-axis-length --> does not fail
            "absolute": takes the absolute values --> may fail
        
        :returns: cropped_data, cropped_ppmscale
        """
        
        if (xmin is not None) or (xmax is not None):

            if mode == "percent":
                ppmmin = xmin/100*self.data[0, :].size
                ppmmax = xmax/100*self.data[0, :].size
                if original_data:
                    self.data = self.data_orig[:, int(ppmmin):int(ppmmax)]
                else:
                    self.data = self.data[:, int(ppmmin):int(ppmmax)]
                self.ppm = self._ppmscale[int(ppmmin):int(ppmmax)]

            elif mode == "absolute":
                if ppm:
                    #print('xmin: %i' % xmin)
                    #print('xmax: %i' % xmax)
                    xmax = np.where(self._ppmscale <= xmax)[0][-1]  # if xmax is not None else -1
                    xmin = np.where(self._ppmscale >= xmin)[0][0]  # if xmin is not None else 0
                    #print('xmin: %i' % xmin)
                    #print('xmax: %i' % xmax)
                    #print(self._ppmscale)
                
                if original_data:
                    self.data = self.data_orig[:, xmin:xmax]
                else:
                    self.data = self.data[:, xmin:xmax]
                    
                self.ppm = self._ppmscale[xmin:xmax]
            else:
                raise ValueError('oops, you mistyped the mode')
            
        return self.data, self.ppm # self._ppmscale[xmin:xmax]
 

    def save_eNMRpy(self, path):
        '''
        saves the instance variables of the object in a .eNMRpy file
        '''
        
        out_dic = {"import class": self.__class__.__name__,}
        
        # reduced number of instance variables for 1D-Measurements which may be potentially saved
        if self.__class__.__name__ == "Measurement":
            instance_variables_keys = ['path', 'expno', 'dateipfad', 'alias', 'name', 'lineb', 'title_page', 'dic', 'data', 'ppm',
                                'pdata', '_ppm_r', '_ppm_l', 'TD', 'n_zf_F2', 'gamma', 'nuc']
        
        # full range of instance variables for an eNMR-Measurement
        else:
            instance_variables_keys = ['path', 'expno', 'dateipfad', 'alias', 'name', 'lineb', 'title_page', 'dic', 'data', 'ppm',
                                'pdata', '_ppm_r', '_ppm_l', 'TD', 'n_zf_F2', 'gamma', 'nuc', 'lin_res_dic', '_x_axis',
                                'dependency', 'cell_resistance', 'Delta', 'delta', 'eNMRraw', 'difflist', 'd', 'g']
        
        # check if the given path ends with .eNMRpy
        if path.split('.')[-1] != 'eNMRpy':
            # if not, change path end to .eNMRpy
            path += '.eNMRpy'
        
        # write selected instance variables in output-dictionary
        for k in instance_variables_keys:
            # if instance variable is np-array
            if type(self.__dict__[k]) == np.ndarray:
                # write array as re-readable string
                s = StringIO()
                np.savetxt(s, self.__dict__[k])
                out_dic[k] = s.getvalue()
                pass
            # write pandas DataFrames in json-format
            elif type(self.__dict__[k]) == pd.DataFrame:
                out_dic[k+'_type_pd.DataFrame'] = self.__dict__[k].to_json()
            # write everything else
            else:
                out_dic[k] = self.__dict__[k]
                
        f = open(path,'w')
        f.write(str(out_dic))
        f.close()
        return



from eNMRpy.Measurement.eNMR_Methods import _eNMR_Methods
import pandas as pd
import numpy as np
from io import StringIO

class Load_eNMRpy_data(_eNMR_Methods):
    '''
    this class is used to load .eNMRpy-files which were saved previously
    
    returns: eNMR-Measurement-object which can be used as usual.
    '''
    def __init__(self, path, load_spectral_data=True):
        
        self.load_spectral_data=load_spectral_data
        
        if path.split('.')[-1] != 'eNMRpy':
            raise NameError('the given path does not correspond to a .eNMRpy file!')
            return
        
        f = open(path)
        
        dic = eval(f.read().replace('array','').replace('matrix', ''))
            
        for k in dic:
            if '_type_pd.DataFrame' in k:
                setattr(self, k.replace('_type_pd.DataFrame',''), pd.read_json(dic[k]))
            elif ((k=='data') or (k=='ppm')) and (load_spectral_data != True):
                pass
            else:
                setattr(self, k, dic[k])
        
        if load_spectral_data:
            self.data = np.loadtxt(StringIO(self.data), dtype=complex)
            self.ppm = np.loadtxt(StringIO(self.ppm), dtype=float)
            
            # derived instance variables
            self.data_orig = self.data
            self._ppmscale = np.linspace(self._ppm_l, self._ppm_r, self.n_zf_F2)  # np.size(self.data[0,:]))
            #self.ppm = self._ppmscale
            self.fid = self.data
            
    def __repr__(self):
        if self.load_spectral_data:
            return '''%s, expno %s, Delta = %.1fms, ppm range: %.1f to %.1f
        delta= %.1fms, g= %.3f T/m, e-distance=%.0fmm'''%(
            self.nuc, self.expno, self.Delta*1000, 
            self.ppm[0], self.ppm[-1],
            self.delta*1000, self.g, self.d*1000
            )
        else:
            return '''spectral data not loaded!
        %s, expno %s, Delta = %.1fms, 
        delta= %.1fms, g= %.3f T/m, e-distance=%.0fmm'''%(
            self.nuc, self.expno, self.Delta*1000, 
            self.delta*1000, self.g, self.d*1000
            )
    
    def plot_fid(self):
        raise ValueError('this method is not available in loaded .eNMRpy file! Sorry! please consider the original data')
    
    def proc(self):
        raise ValueError('this method is not available in loaded .eNMRpy file! Sorry! please consider the original data')

def open_measurement(_cls, path):
    '''
    opens .eNMRpy file in path as an instance of _cls
    '''
    f = open(path)
    s = eval(f.read())
    f.close()
        
    obj =  _cls(**(s))
    obj.__dict__.update(s)

    return obj




def relegend(fig, new_labels, **kwargs):
    '''
    Takes a figure with a legend, gives them new labels (as a list) 
    
    **kwargs: ncol, loc etc.
        ncol: number of columns
        loc: location of the legend (see matplotlib documentation)
    '''
    ax = fig.gca()
    handles, labels = ax.get_legend_handles_labels()
    ax.legend(handles, new_labels, **kwargs) 

def calibrate_x_axis(fig, val, majorticks=1, new_xlim=None):
    """
    this function takes a pyplot figure and shifts the x-axis values by val.
    majorticks defines the distance between the major ticklabels.
    """
    ax = fig.gca()
    xlim = ax.get_xlim()
    
    if xlim[0] > xlim[-1]:
        x1, x2 = ax.get_xlim()[::-1]
    else:
        x1, x2 = ax.get_xlim()
        
    ax.set_xticks(np.arange(x1, x2+1, majorticks)-val%1)
    ax.set_xticklabels(['%.1f'%f for f in ax.get_xticks()+val])
    if new_xlim is None:
        ax.set_xlim(xlim)
    else:
        ax.set_xlim(new_xlim)
        
    return spec

def phc_normalized(phc0, phc1):
    """
    nifty function to correct the zero order phase correction
    when phasing 1st order distortions. 
    phc0 -= phc1/(np.pi/2)
    
    returns: (phc0, phc1)
    
    best use in conjunctoin with self.proc()
   
    """
    
    #phc0 -= phc1/(np.pi/2)
    phc0 -= phc1/(np.pi)
    return phc0, phc1



from .eNMR_Methods import _eNMR_Methods
import pandas as pd
import xml.etree.ElementTree as etree

class Pavel(_eNMR_Methods):
    '''
    This is the subsubclass of Masurement() and subclass of eNMR_Methods specialised to process data obtained from the experimental Swedish from Pavel set-up
    the voltage list is valculated from the vd-values

    path:
        relative or absolute path to the measurements folder
    expno:
        the to the experiment number corresponding EXPNO
    dependency:
        'U': voltage dependent eNMR measurement
        'G': fieldgradient dependent eNMR measurement
             -->reads only the first value in the vd-List to calculate the set voltage

    alias:
        Here you can place an individual name relevant for plotting. If None, the path is taken instead.
    lineb:
        setting a standard-value for the linebroadening.
    d:
        electrode_distance
        
    Uconst:
        constant voltage value set for gradient dependent measurements (dependency="G")
    '''
    def __init__(self, path, expno, dependency='U', alias=None, lineb=.3, d=2.2e-2, cell_resistance=None, Uconst=None):
        
        self.dependency = dependency
        self.cell_resistance = cell_resistance
        
        super().__init__(path, expno, lineb=lineb, alias=alias)
        
        #self._x_axis = {"G": 'g in T/m', "U": 'U / [V]'}[dependency.upper()]
        # import the diffusion parameters
        diffpar = etree.parse(self.dateipfad+'/diff.xml')
        root = diffpar.getroot()
        self.Delta = float(root.findall('DELTA')[0].text)*1e-3
        self.delta = float(root.findall('delta')[0].text)*1e-3  #in Seconds
        print('The diffusion parameters were read from the respective .XML!')
        
        try:
            self.vdList = pd.read_csv(self.dateipfad+"/vdlist",
                                      names=["vd"]).loc[:len(self.data[:, 0])-1]
        except IndexError:
            raise IndexError('Your data is %i-dimensional instead of 2-dimensional. This may be an issue with Topspin. Most likely your measurement was aborted.'%self.data.ndim)
        except:
            raise FileNotFoundError('no VD-List found!')
            #print('no vdList found, generated ones list instead')
            #self.vdList = pd.DataFrame(np.ones((len(self.data[:, 0]), 1)),
                                       #columns=["vd"])
        self.eNMRraw = self.vdList

        #self.vdList["U / [V]"] = hier die Konversion von vdlist zu Spannungsliste
        try:
            self.difflist = pd.read_csv(self.dateipfad+"/gradlist",
                                    names=["g in T/m"])*0.01
        except:
            print('gradlist not found. difflist imported instead')
            self.difflist = pd.read_csv(self.dateipfad+"/difflist",
                                    names=["g in T/m"])*0.01
        self.eNMRraw["g in T/m"] = self.difflist
        
        self.d = d
        self.g = self.eNMRraw["g in T/m"][0]
        

        if self.dependency.upper() == "G":
            self.uInk = None
            # was introduced for vdlists with only a single entry but works regardless of the number of entries if dependency=='G', since the voltage is kept constant
            #if self.vdList.shape[0] == 1: 
            self.eNMRraw = self.difflist
            self.eNMRraw['vd'] = float(self.vdList.iloc[0, 0][:-1])
            self.eNMRraw['U / [V]'] = [
            0 if (self.eNMRraw.loc[i,'vd'] <= 0.6)
            else 
            n 
            for i, n in enumerate(self.eNMRraw['vd']*5)]
            
        else:
            # converts the vd-List
            for i, n in enumerate(self.eNMRraw['vd']):
                self.eNMRraw.loc[i, 'vd_temp'] = float(n[:-1])
            # calculates the applied Voltages        
        
        if (self.dependency.upper() == "U") and (self.dic['acqus']['PULPROG'] == 'py_ENMR_DSTE_3.2_AL'): # checks additionally for the used PULPROG
            self.eNMRraw[self._x_axis] = [
                0 if (self.eNMRraw.loc[i,'vd_temp'] <= 0.6)
                else 
                n if i%2==0 
                else 
                n*-1 # voltage reversal inherent to the PULPROG
                for i, n in enumerate(self.eNMRraw['vd_temp']*5)]
        
            self.uInk = self.eNMRraw['U / [V]'][0] - self.eNMRraw['U / [V]'][1] 
            if self.uInk == 0:
                self.uInk = self.eNMRraw['U / [V]'][0] - self.eNMRraw['U / [V]'][2] 
            if self.uInk < 0:
                self.uInk *= -1
        
        elif self.dependency.upper() == "I":
            self.uInk = None
            self.eNMRraw[self._x_axis] = [
                    0 if (self.eNMRraw.loc[i,'vd_temp'] <= 0.6)
                    else 
                    n if i%2==0 
                    else 
                    n*-1
                    for i, n in enumerate(self.eNMRraw['vd_temp'])
                ]
        
        elif self.dependency.upper() == "RI":
            self.uInk = None
            self.eNMRraw[self._x_axis] = [
                    0 if (self.eNMRraw.loc[i,'vd_temp'] <= 0.6)
                    else 
                    n if i%2==0 
                    else 
                    n*-1
                    for i, n in enumerate(self.eNMRraw['vd_temp'])
                ]
            self.uInk = self.eNMRraw['RI / V'][0] - self.eNMRraw['RI / V'][1] 
            if self.uInk == 0:
                self.uInk = self.eNMRraw['RI / V'][0] - self.eNMRraw['RI / V'][2] 
            if self.uInk < 0:
                self.uInk *= -1
            # calculation of the Voltage from cell resistance and Current /1000 because of mA
            self.eNMRraw[self._x_axis] *= self.cell_resistance/1000

         
    def plot_spec(self, row, xlim=None, figsize=None, invert_xaxis=True, sharey=True):#, ppm=True):
        from .Juergen1 import Juergen1 as eNMR_Measurement
        return eNMR_Measurement.plot_spec(self, row, xlim, figsize, invert_xaxis, sharey)#, ppm=True):



from .eNMR_Methods import _eNMR_Methods
import pandas as pd
import xml.etree.ElementTree as etree
import re
import numpy as np


class Flo(_eNMR_Methods):
    '''
    This is the subsubclass of Masurement() and subclass of eNMR_Methods specialised to process data obtained from the experimental Swedish from Pavel set-up
    the voltage list is valculated from the vd-values

    path:
        relative or absolute path to the measurements folder
    expno:
        the to the experiment number corresponding EXPNO
    dependency:
        'U': voltage dependent eNMR measurement
        'G': fieldgradient dependent eNMR measurement

    alias:
        Here you can place an individual name relevant for plotting. If None, the path is taken instead.
    lineb:
        setting a standard-value for the linebroadening.
    d:
        electrode_distance
    '''
    def __init__(self, path, expno, dependency='U', alias=None, lineb=.3, d=2.2e-2, cell_resistance=None):
        
        self.dependency = dependency
        self.cell_resistance = cell_resistance
        
        super().__init__(path, expno, lineb=lineb, alias=alias)
        
        #self._x_axis = {"G": 'g in T/m', "U": 'U / [V]'}[dependency.upper()]
        # import the diffusion parameters
        diffpar = etree.parse(self.dateipfad+'/diff.xml')
        root = diffpar.getroot()
        self.Delta = float(root.findall('DELTA')[0].text)*1e-3
        self.delta = float(root.findall('delta')[0].text)*1e-3  #in Seconds
        print('The diffusion parameters were read from the respective .XML!')

        if path[-1] == '/':
            pulseprogram = open(path+'pulseprogram')
        else:
            pulseprogram = open(path+'/pulseprogram')
            
        self.pulseprogram = pulseprogram.read()

        bitregex = r"(define list.*bit.*|define list.*pol.*)"
        vlistregex = r".*Voltage\sList.*=.*"

        # list, reading all lines with bit lists in the pulse program
        rawlist = re.findall(bitregex, self.pulseprogram)
        rawvlist = re.findall(vlistregex, self.pulseprogram) 

        rawvlist = rawvlist[0].split('=')
        vlist = eval(rawvlist[1])

        # array of integers generated from the rawlist. Indexing like [bit, voltagestep]
        bitarray = np.array([[int(i) for i in re.findall('{.*}', rawlist[j])[0][1:-1].split(' ')] for j in range(len(rawlist))])


        def byte_to_int(bitarray, row):
            #converting the array into the correct string
            bitstring = str(bitarray[:,row])[1:-1].replace(' ', '')
            
            # check the last bit for the polarity
            if bitstring[-1] == '1':
                polarity = 1
            elif bitstring[-1] == '0':
                polarity = -1
            # transformation of the bitstring minus polarity, which is the last bit, with a base of 2
            intvar = int(bitstring[:-1], 2)
            
            return polarity*intvar


        ulist = [byte_to_int(bitarray, i) for i in range(len(bitarray[0]))]
        
        if ulist == vlist:
            pass
        else:
            raise ValueError('The decoded voltage list does not match the endcoding voltage list! Revisit your pulse program!')
        
        self.eNMRraw = pd.DataFrame(ulist, columns=['U / [V]'])

        try:
            self.difflist = pd.read_csv(self.dateipfad+"/gradlist",
                                    names=["g in T/m"])*0.01
        except:
            print('gradlist not found. difflist imported instead')
            self.difflist = pd.read_csv(self.dateipfad+"/difflist",
                                    names=["g in T/m"])*0.01

        self.eNMRraw["g in T/m"] = self.difflist
        
        self.d = d
        self.g = self.eNMRraw["g in T/m"][0]
        

        ## converts the vd-List
        #for i, n in enumerate(self.eNMRraw['vd']):
            #self.eNMRraw.loc[i, 'vd_temp'] = float(n[:-1])
        ## calculates the applied Voltages

        #if self.dependency.upper() == "U":
            #self.eNMRraw[self._x_axis] = [
                #0 if (self.eNMRraw.loc[i,'vd_temp'] <= 0.6)
                #else 
                #n if i%2==0 
                #else 
                #n*-1
                #for i, n in enumerate(self.eNMRraw['vd_temp']*5)]
        
            #self.uInk = self.eNMRraw['U / [V]'][0] - self.eNMRraw['U / [V]'][1] 
            #if self.uInk == 0:
                #self.uInk = self.eNMRraw['U / [V]'][0] - self.eNMRraw['U / [V]'][2] 
            #if self.uInk < 0:
                #self.uInk *= -1
        
        #elif self.dependency.upper() == "I":
            #self.uInk = None
            #self.eNMRraw[self._x_axis] = [
                    #0 if (self.eNMRraw.loc[i,'vd_temp'] <= 0.6)
                    #else 
                    #n if i%2==0 
                    #else 
                    #n*-1
                    #for i, n in enumerate(self.eNMRraw['vd_temp'])
                #]
        
        #elif self.dependency.upper() == "RI":
            #self.uInk = None
            #self.eNMRraw[self._x_axis] = [
                    #0 if (self.eNMRraw.loc[i,'vd_temp'] <= 0.6)
                    #else 
                    #n if i%2==0 
                    #else 
                    #n*-1
                    #for i, n in enumerate(self.eNMRraw['vd_temp'])
                #]
            #self.uInk = self.eNMRraw['RI / V'][0] - self.eNMRraw['RI / V'][1] 
            #if self.uInk == 0:
                #self.uInk = self.eNMRraw['RI / V'][0] - self.eNMRraw['RI / V'][2] 
            #if self.uInk < 0:
                #self.uInk *= -1
            ## calculation of the Voltage from cell resistance and Current /1000 because of mA
            #self.eNMRraw[self._x_axis] *= self.cell_resistance/1000
            
    def plot_spec(self, row, xlim=None, figsize=None, invert_xaxis=True, sharey=True):#, ppm=True):
        from .Juergen1 import Juergen1 as eNMR_Measurement
        return eNMR_Measurement.plot_spec(self, row, xlim, figsize, invert_xaxis, sharey)#, ppm=True):



from setuptools import setup

setup(
   name='eNMRpy',
   version='0.0.5',
   author='Florian Ackermann',
   author_email='nairolf.ackermann@gmail.com',
   packages=['eNMRpy', 'eNMRpy.Measurement'],
   url='https://github.com/Flackermann/eNMRpy/',
   license='LICENSE.txt',
   description='nmrglue-based package for the import and analysis of electrophoretic NMR-data',
   long_description=open('README.md').read(),
   install_requires=[
       "nmrglue",
       "lmfit>=1.0.1",
       "pandas",
       "matplotlib>=3.1.3",
       "numpy",
       "scipy",
       'scikit-learn<=0.23',
   ],
)



from .eNMR_Methods import _eNMR_Methods
import matplotlib.pyplot as plt
from .base import Measurement
from re import findall
import pandas as pd
import numpy as np

class Juergen1(_eNMR_Methods):
    '''
    This is the subsubclass of Masurement() and subclass of eNMR_Methods specialised to process data obtained from the experimental Schönhoff set-up
    
    path:
        relative or absolute path to the measurements folder
    measurement:
        the to the experiment corresponding EXPNO
    alias:
        Here you can place an individual name relevant for plotting. If None, the path is taken instead.    
    Uink:
        voltage increment. Usually extracted from the title file if defined with e.g. "Uink = 10V"
        If Uink cannot be found or is wrong it can be entered manually during the data import.
        The voltage list is calculated from the voltage increment and the vc list when the incrementation loop is used in the pulse program
    dependency:
        'U': voltage dependent eNMR measurement
        'G': fieldgradient dependent eNMR measurement

    linebroadening:
        setting a standard-value for the linebroadening.
    '''
    def __init__(self, path, expno, Uink=None, dependency="U", alias=None, linebroadening=0.5, electrode_distance=2.2e-2):
        Measurement.__init__(self, path, expno, lineb=linebroadening, alias=alias)
        self.dependency = dependency.upper()
        
        self._x_axis = {"U": "U / [V]",
                   "G": "g in T/m",
                   "I": "I / mA",
                   'RI': 'RI / V'
                   }[self.dependency.upper()]
        
        #self._x_axis = {"G": "g in T/m",
                        #"U": "U / [V]"}[self.dependency.upper()]
        
        if dependency.upper() == "U":
            try:
                # takes the title page to extract the volt increment
                title = open(self.dateipfad+"/pdata/1/title").read()
                # gets the voltage increment using a regular expression
                #uimport = findall('[U|u]in[k|c]\s*=?\s*\d+', title)[0]
                uimport = findall('[U|u]in[k|c]\s*=+\s*\d+', title)[0]
                self.uInk = int(findall('\d+', uimport)[0])
            except ValueError:
                print('no volt increment found\nyou may want to put it in manually')
                self.uInk = Uink
            except IndexError:
                print('No Uink found! May not be an eNMR experiment.')
                self.uInk = Uink
                
        elif dependency.upper() == "G":
            try:
                # takes the title page to extract the volt increment
                title = open(self.dateipfad+"/pdata/1/title").read()
                # gets the voltage increment using a regular expression
                uimport = findall('[U|u]\s*=?\s*\d+', title)[0]
                self.uInk = int(findall('\d+', uimport)[0])

            except ValueError:
                print('no volt increment found\nyou may want to put it in manually')
                self.uInk = Uink
            except IndexError:
                print('No Uink found! May not be an eNMR experiment.')
                self.uInk = Uink # Uinktext

        if self.dependency.upper() == "U":
            try:
                self.vcList = pd.read_csv(self.dateipfad+"/vclist",
                                          names=["vc"]).loc[:len(self.data[:, 0])-1]
            except:
                print("There is a Problem with the VC-list or you performed a gradient dependent measurement")
        elif self.dependency.upper() == "G":
            self.vcList = pd.DataFrame(np.ones((len(self.data[:, 0]), 1)),
                                       columns=["vc"])
        else:
            print("The dependency is not properly selected, try again!")

        self.difflist = pd.read_csv(self.dateipfad+"/difflist",
                                    names=["g in T/m"])*0.01
        
        if Uink is not None:
            self.uInk = Uink
            
        # this implies alternating polarity in the pulse program
        self.vcList["U / [V]"] = [self.vcList["vc"][n]/2*self.uInk if self.vcList["vc"][n] % 2 == 0
                                  else (self.vcList["vc"][n]+1)/2*self.uInk*-1
                                  for n in range(len(self.data[:, 0]))]
        
        # try to open phase data, otherwise create new
        try:
            self.eNMRraw = pd.read_csv(self.path+"phase_data_"+self.expno+".csv",
                                       index_col=0, sep=" ")
            # --> update voltage list
            self.eNMRraw["U / [V]"] = self.vcList["U / [V]"]
        except:
            print("eNMRraw was missing and is generated")
            self.vcList["ph0"] = np.zeros(len(self.data.real[:, 0]))
            self.eNMRraw = self.vcList
        finally:
            self.eNMRraw["g in T/m"] = self.difflist
        
        self.p1 = self.dic["acqus"]["P"][1]
        self.d1 = self.dic["acqus"]["D"][1]
        
        try:
            # import of diffusion parameters for newer Spectrometers
            import xml.etree.ElementTree as etree
            diffpar = etree.parse(self.dateipfad+'/diff.xml')
            root = diffpar.getroot()
            self.Delta = float(root.findall('DELTA')[0].text)*1e-3
            self.delta = float(root.findall('delta')[0].text)*1e-3  # it should be read as in microseconds at this point due to bruker syntax
            print('The diffusion parameters were read from the respective .XML!')
        except:
            # determination of the diffusion parameters for the Spectrometer "Emma"
            _d2 = self.dic["acqus"]["D"][2]
            _d5 = self.dic["acqus"]["D"][5]
            _d9 = self.dic["acqus"]["D"][9]
            _d11 = self.dic["acqus"]["D"][11]
            _p19, _p18, _p17 = self.dic["acqus"]["P"][19], self.dic["acqus"]["P"][18], self.dic["acqus"]["P"][17]
            print('Your data is from an old spectrometer!')
            # calculating usable parameters
            self.delta = _p17+_p18
            _Delta_1 = 0.001*(_p17*2+_p18)+(_d2+_d9+_d5+_d11)*1000+0.001*self.p1+_d11
            _Delta_2 = 0.001*(_p17*2+_p18)+(_d2+_d9+_d5+_d11)*1000+0.001*self.p1*2
            _spoiler = (_d11+_p17+_p19+_p17)*0.001+_d2*1000
            self.Delta = _Delta_1+_Delta_2+2*_spoiler
            self.Delta *=1e-3 # unit conversion
            self.delta *=1e-6 # final unit conversion
            

        # electrode distance in m
        self.d = electrode_distance
        self.g = self.eNMRraw["g in T/m"][0]
    
    def plot_spec(self, row, xlim=None, figsize=None, invert_xaxis=True, sharey=True):#, ppm=True):
        """
        plots row 0 and row n in the range of xmax and xmin
        
        :returns: figure
        """

        _max = None if xlim is None else xlim[0]
        _min = None if xlim is None else xlim[1]
        
        if type(xlim) is not list:
            fig = plt.figure(figsize=figsize)
            ax = fig.add_subplot(111)
        elif type(xlim) == list:
            fig, ax = plt.subplots(ncols=len(xlim), nrows=1, figsize=figsize, sharey=sharey)
        
        _min = self.ppm[0] if xlim is None else xlim[1]
        _max = self.ppm[-1] if xlim is None else xlim[0]
        
        def plot(r, axes=ax):
            
            unit = {'U': 'V', 'G': 'T/m', 'I': 'mA', 'RI': 'V'}[self.dependency.upper()]
            
            axes.plot(self.ppm, self.data[r, ::1].real, label='row %i, %i %s'%(r, self.eNMRraw[self._x_axis].iloc[r], unit))
        
        if type(xlim) is not list:
            if type(row) ==list:
                for r in row:
                    plot(r)
            else:
                plot(row)
        
            ax.set_xlim(xlim)
            #ax.set_title("this is row %i at %.0f V" % (row, self.eNMRraw.loc[row, "U / [V]"]))
            ax.set_xlabel("$\delta$ / ppm")
            ax.set_ylabel("intensity / a.u.")

            ax.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
            if invert_xaxis:
                xlimits = ax.get_xlim()
                ax.set_xlim(xlimits[::-1])
            
            ax.legend()
        
        elif type(xlim) == list:
            for axis, xlim in zip(ax,xlim):
                if type(row) ==list:
                    for r in row:
                        plot(r, axis)
                else:
                    plot(row, axis)
                
                axis.set_xlim(xlim)
                #ax.set_title("this is row %i at %.0f V" % (row, self.eNMRraw.loc[row, "U / [V]"]))
                axis.set_xlabel("$\delta$ / ppm")
                axis.set_ylabel("intensity / a.u.")
                axis.legend()
                axis.ticklabel_format(style='sci', axis='y', scilimits=(0, 0))
                if invert_xaxis:
                    xlimits = axis.get_xlim()
                    axis.set_xlim(xlimits[::-1])
            ax[-1].legend()
            #fig.legend()
        return fig
    
 



"""
Here, you can find the SpecModel Class for creating a fit model to analyse the eNMR spectra.
You can also use the wrapper function set_peaks() to create a fit model from a matplotlib-based GUI input

You can also Simulate Spectra with Phase shifts.
"""

import numpy as np
import matplotlib.pyplot as plt
import lmfit
import pandas as pd
from collections import OrderedDict


def set_peaks(m, n=-1, timeout=-1, xlim=None, **plot_kwargs): #,xlim=(0,10)
    """
    m : measurement object
    
    returns: array of tuples with [(v0, a0), ...] for make_model()
    ______________________________________________________________
    should be used in jupyter notebook like this:
    
    %matplotlib
    peaks = set_peaks(m)
    %matplotlib inline
    make_model(peaks)    
    """
    
    #use built-in method to create pyplot-figure
    m.plot_spec([0], xlim=xlim, **plot_kwargs)
    # adjust the layout for the GUI to minimize bezels
    plt.tight_layout() 
    # insert hints how to correctly use the GUI input
    plt.text(0.95,0.95,'left mouse button: set coordinate \nright mouse button: delete last coordinate\nmiddle mouse button: exit\n',
         horizontalalignment='right',verticalalignment='top', transform=plt.gca().transAxes)
    # activates the input-function delivered by matplotlib
    arr = plt.ginput(n=n, timeout=timeout)
    #returns an array with the selected coordinates
    return arr


def peakpicker(x, y=None, inverse_order=False, width=10, threshold=1e5):
    """
    tool to automatically determine peak positions in an x-y spectrum
    
    x: ppm/frequency-array
        x [optional]: Measurement-object of which the first slice is taken
    y: intensity
    
    inverse_order:
        inverts the order of peaks in the array
        this could be helpful when passing it to make_model
    
    width: scan width. should be even numbered
    
    
    returns: 2D-np.array with [[ppm,intensity],[...]] which can be passed to make_model
    """
    
    #res: step size for scan
    res=1

    if type(x) != np.ndarray:
        y = x.data[0]
        x = x.ppm

    
    if width%2 != 0:
        raise ValueError('width should be an even integer')
#         return
    
    def advindx(arr, min, max):
        """more or less introduces repeating boundaries for the array"""
        if (min < 0) or (max > len(arr)):
            a = arr[min:]
            b = arr[0:max]
            return np.append(a, b)
        else:
            return arr[min:max]
    
    def check_peak(y_out, width=width):
        '''checks for consecutive increase and decrease of values'''
        
        arr = y_out
        
        sel_arr = np.array([])
        for i in range(width):
            # if adjacent values increase, the difference is positive.
            # if adjacent values increase, the difference is negative.
            sel_arr = np.append(sel_arr, (arr[i+1].real - arr[i].real))
        
        # check for increase or decrease
        bool_arr = sel_arr > 0
        
        # if in the first part all values increase and the second part all values decrease
        if (all(bool_arr[:width//2]) == True) and all(not x for x in bool_arr[width//2:]):
            return True
        else:
            return False
    
    peaks = []
    for i in range(0,len(y), res):
        x_out = advindx(x,i-width,i+width)
        y_out = advindx(y,i-width,i+width)

        check = check_peak(y_out, width)
    
        if check and (y[i-width//2].real > threshold):
            # i-width//2 ensures that the right coordinates are given for the respective peak
            peaks.append((x[i-width//2], y[i-width//2].real))
    
    if inverse_order:
        return np.array(peaks)[::-1]
    
    elif not inverse_order:
        return np.array(peaks)


def make_model(peaks, print_params=True):
    '''
    returns a SpecModel()-object with parameters generated from set_peaks()-array
    '''
    
    model = SpecModel(len(peaks))
    # sets the chemical shift vX of each peak
    model.set_initial_values(['v%i'%i for i in range(len(peaks))], [i[0] for i in peaks])
    # sets the estimated amplitude as the peak hight divided by 100
    model.set_initial_values(['a%i'%i for i in range(len(peaks))], [i[1]/100 for i in peaks])
    #shows the resulting parameters in output line
    if print_params:
        model.params.pretty_print()
    return model

# maybe should be handled as static method
def reduce_fitted_phases(Measurementobject, SpecModel):
    """
    Static method for the calculation of reduced phase shift values, where the slope is equal to the
    electrophoretic mobility when plotted against the applied voltage.
    
    This is especially helpful for the comparison of phase shifts obtained for different nuclei
    or under different experimental conditions.
    """
    m = Measurementobject
    for k in ['ph%i'%i for i in range(SpecModel.n)]:
        m.eNMRraw[k+'reduced'] = m.eNMRraw[k]*m.d/m.delta/m.Delta/m.g/m.gamma
        m.eNMRraw[k+'reduced'] -= m.eNMRraw.loc[m.eNMRraw['U / [V]']==0, k+'reduced'][0]

def fit_Measurement(obj_M, obj_S, fixed_parameters=None, plot=False, savepath=None, **plot_kwargs):
    '''
    function to fit a the series of voltage dependent spectra contained in a typical eNMR measurement
    
    obj_M: object of the class eNMR_Measurement
    obj_S: object of the class SpecModel
    fixed_parameters: List of parameters to be fixed after the first fit
    
    **plot_kwargs are passed to SpecModel.fit:
        peak_deconvolution=False, parse_complex='abs','real, or 'imag'
    
    '''
    
    i=0

    if fixed_parameters is not None:
        fp = [] # working list of parameter-objects
        
        # iterates the list of parameters
        for k in obj_S.params.keys(): 
            # if the parameter name p[0] matches any string in fixed_parameters
            if any(k == np.array(fixed_parameters)): 
                # append the parameter to the working list
                fp.append(k)                
        
        
        fig = obj_S.fit(obj_M.ppm, obj_M.data[0], **plot_kwargs)
        print('row 0 fitted including fixed_parameters being varied')
        ph_res = obj_S.get_result_values()
        
        for par in ph_res.keys():
            # saves the results from row 0 in the eNMRraw-DataFrame
            obj_M.eNMRraw.at[0, par] = ph_res[par]
        
        
        for p in fp: # fixes all variables listed in fixed_parameters
            obj_S.params[p].set(obj_S.result.params[p].value)
            obj_S.params[p].set(vary=False)
            print('%s not varied!'%p)
            
        if (plot is True) and (savepath is not None):
            fig.savefig(savepath+'%.1f'%obj_M.eNMRraw.loc[0, obj_M._x_axis]+'.png', dpi=300)
        
        i = 1 #counter set to one for the rest of the spectra to be fitted

    print('start fitting from row %i'%i)
    
    for row in range(i, obj_M.data[:,0].size):
        fig = obj_S.fit(obj_M.ppm, obj_M.data[row], plot=plot, **plot_kwargs)
        ph_res = obj_S.get_result_values()
        
        for par in ph_res.keys():
            #obj_M.eNMRraw.set_value(row, par, ph_res[par])
            obj_M.eNMRraw.at[row, par] = ph_res[par]
            
        if (plot is True) and (savepath is not None):
            fig.savefig(savepath+'%.1f'%obj_M.eNMRraw.loc[row, obj_M._x_axis]+'.png', dpi=300)
            
    #for p in fp: # reset all vary-Values
        #obj_S.params[p].set(vary=True)
    
    
    print('fitting finished')

def drop_errors(df):
    '''
    drops all columns that which keys end with _err --> created from the fitting model
    This function is used in the plot_correlations function
    '''

    df.keys()
    #drop the vc, Voltag, Gradient and outlier columns
    try:
        sel = df.drop(['vc', 'U / [V]', 'g in T/m', 'outlier'], axis=1)
    except KeyError:
        try:
            sel = df.drop(['vd', 'U / [V]', 'g in T/m', 'outlier'], axis=1)
        except:
            print('no vd or vc found, no standard parameters dropped')
            sel = df
            #sel = df.drop(['vd', 'U / [V]', 'g in T/m', 'outlier'], axis=1)
    # drop all error-columns-boolean
    _bool = np.array([k[-4:] != '_err' for k in sel.keys()])
    # elect only the non-error columns
    sel = sel[np.array(sel.keys()[_bool])]
    return sel

def plot_correlations_heatmap(df, method='pearson', without_errors=True, textcolor="#222222", **fig_kwargs):
    """
    correlation coefficients plot (heatmap) for any pandas DataFrame
    
    method:
        pearson, kendall, spearman
    """
    if without_errors:
        df = drop_errors(df)
    corr = df.corr(method=method)
    columns = corr.keys()
    indices = np.array(corr.index)

    fig, ax = plt.subplots(**fig_kwargs)
    im = ax.imshow(corr, cmap='Spectral', vmin=-1, vmax=1)

    # We want to show all ticks...
    ax.set_xticks(np.arange(len(columns)))
    ax.set_yticks(np.arange(len(indices)))
    # ... and label them with the respective list entries
    ax.set_xticklabels(columns)
    ax.set_yticklabels(indices)

    # Rotate the tick labels and set their alignment.
    plt.setp(ax.get_xticklabels(), rotation=45, ha="right",
             rotation_mode="anchor")

    # Loop over data dimensions and create text annotations.
    for i in range(len(indices)):
        for j in range(len(columns)):
            text = ax.text(j, i, '%.3f'%corr.iloc[i, j],
                           ha="center", va="center", color=textcolor)
    plt.colorbar(im)
    fig.tight_layout()
    
    return fig, corr


def lorentz_real(x, x0, _lambda, ph, amplitude):
    """
    calculates the real part of the spectrum
    here, the conversion from rad to ° happens
    """
    def dispersion(x, x0, _lambda):
        return -(x-x0)/(_lambda**2+(x-x0)**2)

    def absorption(x, x0, _lambda):
        return _lambda/(_lambda**2+(x-x0)**2)
    # transforming the phase angle from degree ° in rad
    ph = (ph/360*2*np.pi)
    amplitude, _lambda = abs(amplitude), abs(_lambda)

    return amplitude*(absorption(x, x0, _lambda)*np.cos(ph)-dispersion(x, x0, _lambda)*np.sin(ph))

def lorentz_imag(x, x0, _lambda, ph, amplitude):
    """
    calculates the imaginary part of the spectrum
    here, the conversion from rad to ° happens
    """
    def dispersion(x, x0, _lambda):
        return -(x-x0)/(_lambda**2+(x-x0)**2)

    def absorption(x, x0, _lambda):
        return _lambda/(_lambda**2+(x-x0)**2)
    
    # transforming the phase angle from degree ° in rad
    ph = (ph/360*2*np.pi)
    amplitude, _lambda = abs(amplitude), abs(_lambda)

    return amplitude*(dispersion(x, x0, _lambda)*np.cos(ph)+absorption(x, x0, _lambda)*np.sin(ph))

def gauss(x, x0, s, amp=1):
    """
    this is just the function of a gaußian distribution with x0 as the center value, s the standard deviation and
    amp as an amplitude which would be 1 in the case of a normal distribution.
    """
    return amp*np.exp(-(x-x0)**2/(2*s**2))/(np.sqrt(np.pi*2)*s)

def makefunc_Lorentz_cmplx(n = 1):
    '''
    returns a function describing the complex lorentzians with real and imaginary part
    n: number of lorentzians contained in the function
    '''
    s = 'lambda x, baseline'
    for i in range(n):
        s += ', v%i, l%i, ph%i, a%i'%(i, i, i, i)
    
    s+=': baseline'
    for i in range(n):
        s += ' + lorentz_real(x, v%i, l%i, ph%i, a%i)'%(i, i, i, i)
        s += ' + 1j*lorentz_imag(x, v%i, l%i, ph%i, a%i)'%(i, i, i, i)
    
    func = eval(s)
    func.__name__="Lorentz Peak Superposition"
    return func

def makefunc_Voigt(n = 1):
    '''
    returns a combination of n Voigt-real as a lambda function
    
    similar to makefunc_Lorentz_cmplx, but with voigt profiles
    '''
    s = 'lambda x, baseline'
    for i in range(n):
        s += ', v%i, l%i, ph%i, a%i, s%i'%(i, i, i, i, i)
    
    s+=': baseline'
    for i in range(n):
        s += ' + lorentz_real(x, v%i, l%i, ph%i, a%i)*gauss(x, v%i, s%i, 1)'%(i, i, i, i, i, i)
        s += ' + 1j*lorentz_imag(x, v%i, l%i, ph%i, a%i)*gauss(x, v%i, s%i, 1)'%(i, i, i, i, i, i)
    
    func = eval(s)
    func.__name__="Voigt Peak Superposition"
    return func

class SpecModel(object):
    '''
    This class creates a lmfit Model from s function of n_peaks lorentz-distributions
    parameter explanation
        an = amplitude n
        ln = peak broadness of n
        phn = phase angle n
        vn = chemical shift v of n
        baseline = basline of the whole spectrum
        model = 'Lorentz' or 'Voigt'
    '''
    
    def __init__(self, n_peaks, verbose=False, model='Lorentz'):
        
        self.mkey = model
        _func = {'Lorentz': makefunc_Lorentz_cmplx,
                 'Voigt' : makefunc_Voigt}[model]
        
        # number of peaks in the fitmodel
        self.n = n_peaks
        
        # the chosen model type is parsed with the number of peaks as a the respective function
        # to the lmfit.Model-wrapper
        self.model = lmfit.Model(_func(n_peaks))
        # parameter set is create
        self.params = self.model.make_params(verbose=False)
        
        self.result = None
        
        # This loop initializes all parameters by 1 and sets standard boundaries depending on the respective parameterset
        for k in self.params.keys():
            self.set_initial_values(k, 1)
            if k[0] == 'a':
                self.set_boundaries(k, 0, np.inf)
                self.set_initial_values(k, 1e6)
            elif k[0] == 'l':
                self.set_boundaries(k, 0, np.inf)
                self.set_initial_values(k, 1e-2)
            elif k[0] == 'p':
                self.set_boundaries(k, -180, 180)
            elif k[0] =='s':
                self.set_boundaries(k, 0, np.inf)
                self.set_initial_values(k, 1e-2)
            else:
                pass
        
        # should this be depreciated?
        if verbose:
            self.params.pretty_print()

    def set_mathematical_constraints(self, expr=None, reset=True):
        """
        expr: mathematical expression without whitespace
        
        reset: will reset all mathematical constraints
        
        example 1:
            ph0 should be always equal to ph1
            set_mathematical_constraint('ph0=ph1')
        example 2:
            ph0 should be always ph1 - 90 degree
            set_mathematical_constraint('ph0=ph1-90')
        """
        
        if (expr == None) and (reset):
            for i in self.params:
                self.params[i].expr = None
                self.params[i].vary = True
            print('model without constraints')
            return
        
        elif reset:
            for i in self.params:
                self.params[i].expr = None
                self.params[i].vary = True
            print('constraints were reset before new assignment')
        
        if type(expr) == list:
            for e in expr:
                a = e.split('=')
                self.params[a[0]].expr = a[1]
        else:
            a = expr.split('=')
            self.params[a[0]].expr = a[1]
        
         
        
    def set_initial_values(self, par, val):
        '''
        takes single parameter par or list of parameters
        to set the initial value as val, which can be a list as well
        '''

        if type(par) == list:
            i = 0
            for p in par:
                self.params[p].value = val[i]
                i+=1
        else:
            self.params[par].value = val
        
    def set_boundaries(self, par, min=None, max=None):
        """
        explicit name for setting min and max of said parameter
        
        par: parameter key for self.params
        """
        self.params[par].min, self.params[par].max = min,max

    
    def calc_single_peak(self, x, params=None, peak=0):
        '''
        calculates the single peak of a parameter set params
        
        mainly used for plotting purposes to visualize the individual peaks fitted in a superposition
        '''
        if params is None:
            params = self.params
        
        def make_singlepeak_params(params, peak=0):
            '''
            creates a dictionary containing all parameters of the respective peak named as 
            0-Peak to pass it to the SpecModel evaluation function of a single peak
            '''
            dic = {k[:-1]+'0': params[k]for k in params.keys() if k[-1] == str(peak)}
            dic['baseline'] = params['baseline']
            return dic

        single_peak = SpecModel(1, model=self.mkey)
        s0par = make_singlepeak_params(params, peak)
        return single_peak.model.eval(s0par, x=x)
        
    def plot_init_spec(self, xdata, single_peaks=False, fig=None):
        '''
        plots the spectrum for the initialized parameter set
        '''
        if fig is None:
            fig, ax = plt.subplots()
        else:
            ax = fig.gca()
        
        if single_peaks:
            for n in range(self.n):
                ax.plot(xdata, self.calc_single_peak(np.array(xdata), params=self.params, peak=n).real, '--', label='peak '+str(n))
        
        ax.plot(np.array(xdata), self.model.eval(x=np.array(xdata), params=self.params).real, c='k', label='start parameters')
        ax.legend()
        
        # get the chemical shifts of all peaks one may also use self.n at this point
        # shiftlist = list(filter(lambda x: x[0]=='v', self.params.keys()))
        for i in range(self.n):
            x = self.params['v%i'%i].value
            # the factor 100 comes from the division in the make_model function, where the amplitude was estimated as y/100
            y = self.params['a%i'%i].value*100 
            # string for the annotation
            s = 'P%i'%i
            ax.annotate(s, (x,y), fontsize=16)
            
        return fig
    
    def fit(self, xdata, ydata, plot=False, peak_deconvolution=False, parse_complex='real', figsize=None):
        """
        method to fit a single spectrum consisting of xdata and ydata
        with the previously defined model and parameters
        
        stores the result in self.result
        
        parse_complex: representation of the complex data when plotting
            "real", "imag", or "abs"
        
        returns a matplotlib figure if plot=True
        """
        
        # use the fit method built in the lmfit model object lmfit
        self.result = self.model.fit(ydata, x=xdata, params=self.params)
        
        fig = None
        if plot:
            fig = self.result.plot(parse_complex=parse_complex)[0]
            if figsize != None:
                fig = plt.gcf()
                fig.set_size_inches(figsize)
            
        if peak_deconvolution and plot:
            ax = fig.gca()
            for n in range(self.n):
                ax.plot(xdata, self.calc_single_peak(np.array(xdata), params=self.result.params, peak=n), '--', label='peak '+str(n))

        return fig
    
    #formerly: def get_phasedata(self):
    def get_result_values(self):
        """
        method to return parameters and their errors resulting from the fit (self.result)
        """
        dic = {}
        #for i in range(self.n):
            #dic['ph%i'%i] = self.result.best_values['ph%i'%i]
            #dic['ph%i_err'%i] = self.result.params['ph%i'%i].stderr
            
        for k in self.params.keys():
            dic[k] = self.result.best_values[k]
            dic[k+'_err'] = self.result.params[k].stderr

        return dic
    
    def report(self):
        """shortcut to print self.result.fit_report()"""
        print(self.result.fit_report())
        
    def reassign_parameter(self, par, best_values_dic, vary=False):
        '''
        reassigns the initial value for the parameter by taking the value from the best-fit dictionary (usually obtained form a test-fit)
        
        vary: sets if the parameter should be varied in further fittings or not.
        '''
        self.params[par].set(best_values_dic[par], vary=vary)
        print(par, '=', best_values_dic[par], ';', 'vary =', vary)


class SpecSim(object):
    '''
    This class lets you simulate peaks with given properties in order compare e.g. MOSY and fitting results to yours if Peaks are overlapping.
    '''
    def __init__(self, n_Peaks, ppm=None, xlim=None, n_points=2**11, model="Lorentz", fit_params=None):
        """
        model: 'Lorentz' or 'Voigt'
        """
        
        _func = {'Lorentz': makefunc_Lorentz_cmplx,
            'Voigt' : makefunc_Voigt}[model]
        self.data = None
        self.func = _func(n_Peaks)
        self.pkeys = self.func.__code__.co_varnames
        self.pdic = OrderedDict({k: 0.1 for k in self.pkeys})
        if fit_params is not None:
            try:
                for p in fit_params.keys():
                    self.pdic[p] = fit_params[p].value
            except:
                print('This fit_params are no lmfit-Parameters object')

        print('func_params', self.pkeys)
        
        if ppm is not None:
            self.ppm = ppm
        elif xlim is not None:
            self.ppm = np.linspace(*xlim, n_points)
        else:
            self.ppm = None
        self.pdic['x'] = self.ppm
        
        self.eNMRraw = pd.DataFrame()
        self.eNMRraw['U / [V]'] = None
        self.params = None

        
    def set_params(self, par, val):
        '''
        sets the parameter par as value value in the parameters dictionary self.pdic
        
        par:
            string or list of strings representing the respective keyword for the parameters
        var: single value or list of values. This can be anything, also numpy arrays.
        '''
        if type(par) == list:
            i=0
            for p in par:
                self.pdic[p] = val[i]
                i += 1
        else:
            self.pdic[par] = val
            
    def add_noise(self, scale=1):
        '''
        adds normal distributed noise to the spectrum
        '''
        self.data += scale*np.random.standard_normal(self.data.shape)
        

    def calc_spec(self, noise=0):
        '''
        calculates the spectrum from the parameters dictionary self.pdic
        '''
        self.data = self.func(*self.pdic.values())
        self.add_noise(noise)
        return self.data
    
    def calc_spec_series(self, params, vlist=None, noise=0):
        '''
        calculates a series of spectra with the given lists.
        if par is a list, then vlist needs to be 2-dimensional and rectangular.
        '''
        if vlist is None:
            par = params.keys()
            vlist = params.values()
            #self.Ulim = params.spec_par['Ulim']

        else:
            vlist = np.array(vlist)
        
        if (vlist.dtype == 'int64') or (vlist.dtype == 'float64'):
            if type(par) != list:
                self.set_params(par, vlist[0])
                self.data = np.array([self.calc_spec(noise)])
                for i in range(1, len(vlist)):
                    self.set_params(par, vlist[i])                
                    self.data = np.append(self.data, np.array([self.calc_spec(noise)]), axis=0)

            elif type(par) == list:
                # set all first parameter values
                for i in range(len(par)):
                    self.set_params(par[i], vlist[i][0])
                # create first row of dataset so rest can be appended in the following loop
                self.data = np.array([self.calc_spec(noise)])
                for j in range(1, len(vlist[0])):
                    for i in range(len(par)):
                        self.set_params(par[i], vlist[i][j])
                    self.data = np.append(self.data, np.array([self.calc_spec(noise)]), axis=0)
        else:
            print('error: vlist is not in a rectangular shape')

        #Instanzvariablen Updaten
        self.params = params
        self.eNMRraw['U / [V]'] = params.U

        _d = pd.DataFrame(self.data)
        self.eNMRraw['data'] = None
        for i in range(len(self.data[:,0])):
            self.eNMRraw.at[i, 'data'] = _d[i]#test[i]
        
        for p in params.keys():
            self.eNMRraw[p] = params[p]
        
        return self.data
    
    def plot(self, text_pos=None):
        fig, ax = plt.subplots()
        if self.data.ndim == 1:
            ax.plot(self.ppm, self.data.real)
        else:    
            for n in range(len(self.data[:,0])):
                ax.plot(self.ppm, self.data[n].real, label=n)
        ax.legend()
        if text_pos is not None:
            ax.text(*text_pos, self.params.spec_par)
        return fig
    
    
class ParameterVariation(OrderedDict):
    '''
    OrderedDictionary with extra methods for the creation of datasets
    '''
    def __init__(self):
        self.spec_par = {'Delta': None,
                         'delta': None,
                         'g': None,
                         'Ulim':None,
                         'NUC':None,
                         }
        self.U = None
        pass
    def values(self):
        return np.array(list(super().values()))
    
    def keys(self):
        return list(super().keys())
    
    def set_par(self, par, val):
        if type(par) != list:
            self[par] = val
        else:
            for i, p in enumerate(par):
                self[p] = val[i]
    
    def set_ph_from_mobility(self, mobilities, n, spec_par=None):
        '''
        mobilities:
            must be an array. mobilities are sorted according
        
        spec_par: dictionary with the spectroscopic parameters in SI units:
            Delta in s
            delta in s
            NUC: 1H, 7Li or 19F
            g in T/m
            Ulim: Voltage limits in V
            d: electrode distance
        '''
        if spec_par is None:
            spec_par = self.spec_par
        
        #gamma in MHz/T
        #gamma = {"7Li": 5957443266,
                #"1H": 15326621020,
                #"19F": 14415618125}[spec_par['NUC']]
        # gamma in rad/Ts
        gamma = {'1H':26.7513e7,
                '7Li': 10.3962e7,
                '19F': 25.1662e7}[spec_par['NUC']]
        # Umrechnung von rad in °
        gamma = gamma/2/np.pi*360
        
        U = np.linspace(*spec_par['Ulim'], n)
        print(U)
        g = spec_par['g']
        d = spec_par['d']
        Delta = spec_par['Delta']
        delta = spec_par['delta']
        
        for i, mu in enumerate(mobilities):
            self['ph%i'%i] = gamma*Delta*delta*g*(U/d)*mu

        self.U = U
        self.spec_par = spec_par







from .base import Measurement
from sklearn.linear_model import huber as hub
import numpy as np
import nmrglue as ng
import matplotlib.pyplot as plt
import pandas as pd
import lmfit as lf

class _eNMR_Methods(Measurement):
    """
    This is the subclass of Masurement() containing all methods
    
    path:
        relative or absolute path to the measurements folder
    measurement:
        the to the experiment corresponding EXPNO
    alias:
        Here you can place an individual name relevant for plotting. If None, the path is taken instead.
    """

    def __init__(self, path, expno, alias=None,  lineb=5):
        super().__init__(path, expno, alias=None,  lineb=5)
        
        self._x_axis = {"U": "U / [V]",
                   "G": "g in T/m",
                   "I": "I / mA",
                   "RI": "RI / V"
                   }[self.dependency.upper()]
    
    def __repr__(self):
        return '''%s, expno %s, Delta = %.1fms, ppm range: %.1f to %.1f
    delta= %.1fms, g= %.3f T/m, e-distance=%.0fmm'''%(
        self.nuc, self.expno, self.Delta*1000, 
        self.ppm[0], self.ppm[-1],
        self.delta*1000, self.g, self.d*1000 # outputs the gradient time and diffusion time in ms
        )
    
    def __getitem__(self, key):
        if type(key) == str:
            return self.eNMRraw[key]
        elif type(key) == int:
            return (self.ppm, self.data[key])                

    def autophase_phase_analysis(self,
                  returns=False,
                  method="acme",
                  progress=False,
                  period_compensation=True,
                  normalize=True):
        """
        analyzes the phase of the spectral data and returns phased data

        at this point no data cropping --> full spectrum is analyzed

        returns: if True, returns the raw phased Data. If false, returns nothing

        method: chooses the method for the phase correction
            "acme": standard method using entropy minimization
            "difference": --> _ps_abs_difference_score()
                minimizes the linear difference to the first spectrum by fitting p0
            "sqdifference": --> _ps_sq_difference_score()
                minimizes the square difference to the first spectrum by fitting p0

        progress:
            prints the progress. In case you are dealing with large datasets

        period_compensation:
            corrects for the 360 degree error that occasionaly occurs since a phase of 0 and 360
            is equivalent and indistinguishable

        normalize:
            normalizes all phase data to 0 degrees for 0V. This compensates for different reference 
            phases during the aquisition and phase analysis.
        """
        if self.dependency.upper() == 'G':
            normalize = False

        data_ph = np.array([])
        for n in range(len(self.data[:, 0])):
            if progress:
                print("row %i / %i" % (n+1, len(self.data[:, 0])))

            #####################################
            # using the data with the algorithm
            _val = self._autops(self.data[n, :], _fnc=method)
            #####################################

            data_ph = np.append(data_ph, _val)
                
            if method == 'acme':
                self.eNMRraw.loc[n, "ph0acme"] = _val[0]  
                if progress:
                    clear_output() # keeps the output clean. remove for more info on iteration steps etc.
            elif method == 'difference':
                self.eNMRraw.loc[n, "ph0diff"] = _val[0]  
                if progress:
                    clear_output() # keeps the output clean. remove for more info on iteration steps etc.
            else:
                self.eNMRraw.loc[n, "ph0sqdiff"] = _val[0]  
                if progress:
                    clear_output() # keeps the output clean. remove for more info on iteration steps etc.
       
        corr_enmr = self.eNMRraw.sort_values(self._x_axis)
        
        if method == 'acme':
            if period_compensation:
                for m in range(corr_enmr["ph0acme"].size):
                    if corr_enmr["ph0acme"].iloc[m]-corr_enmr["ph0acme"].iloc[abs(m-1)] < -300:
                        corr_enmr["ph0acme"].iloc[m] += 360
                    elif corr_enmr["ph0acme"].iloc[m]-corr_enmr["ph0acme"].iloc[abs(m-1)] > 300:
                        corr_enmr["ph0acme"].iloc[m] -= 360
                self.eNMRraw = corr_enmr
        
            if normalize:
                #self.U_0 = corr_enmr[corr_enmr["U / [V]"] == 0]["ph0acme"][0]
                self.U_0 = corr_enmr[corr_enmr[self._x_axis] == 0]["ph0acme"].iloc[0]
                for m in range(corr_enmr["ph0acme"].size):
                    corr_enmr["ph0acme"].iloc[m] -= self.U_0
                self.eNMRraw = corr_enmr

            self.eNMRraw["ph0acmereduced"] = corr_enmr['ph0acme']*self.d/(self.eNMRraw['g in T/m'][0]*self.Delta*self.delta*self.gamma)
        elif method == 'difference':
            if period_compensation:
                for m in range(corr_enmr["ph0diff"].size):
                    if corr_enmr["ph0diff"].iloc[m]-corr_enmr["ph0diff"].iloc[abs(m-1)] < -300:
                        corr_enmr["ph0diff"].iloc[m] += 360
                    elif corr_enmr["ph0diff"].iloc[m]-corr_enmr["ph0diff"].iloc[abs(m-1)] > 300:
                        corr_enmr["ph0diff"].iloc[m] -= 360
                self.eNMRraw = corr_enmr
        
            if normalize:
                self.U_0 = corr_enmr[corr_enmr[self._x_axis] == 0]["ph0diff"][0]
                for m in range(corr_enmr["ph0diff"].size):
                    corr_enmr["ph0diff"].iloc[m] -= self.U_0
                self.eNMRraw = corr_enmr

            self.eNMRraw["ph0diffreduced"] = corr_enmr['ph0diff']*self.d/(self.eNMRraw['g in T/m'][0]*self.Delta*self.delta*self.gamma)
        else:
            if period_compensation:
                for m in range(corr_enmr["ph0sqdiff"].size):
                    if corr_enmr["ph0sqdiff"].iloc[m]-corr_enmr["ph0sqdiff"].iloc[abs(m-1)] < -300:
                        corr_enmr["ph0sqdiff"].iloc[m] += 360
                    elif corr_enmr["ph0sqdiff"].iloc[m]-corr_enmr["ph0sqdiff"].iloc[abs(m-1)] > 300:
                        corr_enmr["ph0sqdiff"].iloc[m] -= 360
                self.eNMRraw = corr_enmr
        
            if normalize:
                self.U_0 = corr_enmr[corr_enmr[self._x_axis] == 0]["ph0sqdiff"][0]
                for m in range(corr_enmr["ph0sqdiff"].size):
                    corr_enmr["ph0sqdiff"].iloc[m] -= self.U_0
                self.eNMRraw = corr_enmr

            self.eNMRraw["ph0sqdiffreduced"] = corr_enmr['ph0sqdiff']*self.d/(self.eNMRraw['g in T/m'][0]*self.Delta*self.delta/1000*self.gamma)

        print("done, all went well")
        
        if returns is True:
            return data_ph  # , data_spec

    def _autops(self, data, _fnc="acme", p0=0.0, p1=None):  # , p1=0.0):
        """
        FS: modified for eNMR purpose. optimizes only p0
        ----------
        Automatic linear phase correction

        Parameters
        ----------
        data : ndarray
            Array of NMR data.
        _fnc : str or function
            Algorithm to use for phase scoring. Built in functions can be
            specified by one of the following strings: "acme", "peak_minima"
        p0 : float
            Initial zero order phase in degrees.
        p1 : float
            Initial first order phase in degrees.

        Returns
        -------
        ndata : ndarray
            Phased NMR data.

        """
        
        from  scipy.optimize import fmin
        
        self._fnc = _fnc
        self._p0 = p0
        self._p1 = p1
        
        if not callable(_fnc):
            self._fnc = {
                # 'peak_minima': _ps_peak_minima_score,
                'acme': self._ps_acme_score,
                'difference': self._ps_abs_difference_score,
                'sqdifference': self._ps_sq_difference_score
            }[self._fnc]
        
        # this is the behavior for the analysis in the eNMR context
        if p1 == None:
            # self.opt ist die optimierte Phase für das Spektrum.
            self.opt = self._p0  # [self._p0, self.p1]
            self.opt = fmin(self._fnc, x0=self.opt, args=(data, ), disp=0)

            # self.phasedspc = ng.proc_base.ps(data, p0=self.opt[0], p1=self.opt[1])

            return self.opt  # self.phasedspc, self.opt
        
        # this is used to autophase the spectrum and get the values zero and first order
        elif p1 != None:
            # self.opt ist die optimierte Phase für das Spektrum.
            self.opt = (self._p0, self._p1)
            self.opt = fmin(self._fnc, x0=self.opt, args=(data, ), disp=0)

            #self.phasedspc = ng.proc_base.ps(data, p0=self.opt[0], p1=self.opt[1])
            return self.opt  # self.phasedspc, self.opt

    @staticmethod
    def _ps_acme_score(ph, data):
        """
        FS: modified for eNMR purpose. optimizes only p0
        ------------
        Phase correction using ACME algorithm by Chen Li et al.
        Journal of Magnetic Resonance 158 (2002) 164-168

        using only the first derivative for the entropy

        Parameters
        ----------
        pd : tuple
            Current p0 and p1 values
        data : ndarray
            Array of NMR data.

        Returns
        -------
        score : float
            Value of the objective function (phase score)

        """
        stepsize = 1
        

        if (len(ph) == 1):
            s0 = ng.proc_base.ps(data, p0=-ph, p1=0)  # , p1=phc1 --> p1=0 lets the algorithm optimize only p0
        
        else:
            s0 = ng.proc_base.ps(data, p0=-ph[0], p1=ph[1])  # optimizing for ph0 and ph1
        
        data = np.real(s0)
        

        # Calculation of first derivatives --> hier wird das absolute Spektrum erzeugt
        ds1 = np.abs((data[1:]-data[:-1]) / (stepsize*2))
        p1 = ds1 / np.sum(ds1)

        # Calculation of entropy
        p1[p1 == 0] = 1  # was macht diese Zeile? das Verstehe ich noch nicht richtig

        h1 = -p1 * np.log(p1)
        h1s = np.sum(h1)

        # Calculation of penalty
        pfun = 0.0
        as_ = data - np.abs(data)
        sumas = np.sum(as_)

        if sumas < 0:
            pfun = pfun + np.sum((as_/2) ** 2)

        p = 1000 * pfun

        return h1s + p
    
    def _ps_abs_difference_score(self, ph, data):
        """
        FS: modified for eNMR purpose. optimizes only p0
        ------------
        Parameters
        ----------
        pd : tuple
            Current p0 and p1 values
        data : ndarray
            Array of NMR data.

        Returns
        -------
        score : float
            Value of the objective function (phase score)
        """
       
        s0 = ng.proc_base.ps(data, p0=-ph, p1=0)  # , p1=phc1 --> p1=0 lets the algorithm optimize only p0
        phasedspec = np.real(s0)

        penalty = np.sum(np.abs(self.data[0, :].real - phasedspec))
        
        return penalty
    
    def _ps_sq_difference_score(self, ph, data):
        """
        FS: modified for eNMR purpose. optimizes only p0
        ------------
        Parameters
        ----------
        pd : tuple
            Current p0 and p1 values
        data : ndarray
            Array of NMR data.

        Returns
        -------
        score : float
            Value of the objective function (phase score)
        """
       
        s0 = ng.proc_base.ps(data, p0=-ph, p1=0)  # , p1=phc1 --> p1=0 lets the algorithm optimize only p0
        phasedspec = np.real(s0)

        penalty = np.sum(np.square(self.data[0,:].real - phasedspec))  # *1/n-1 wäre die Varianz,
        # ist hier egal, da alle Spektren die gleiche Anzahl an Punkten haben.
        
        return penalty
    
    def analyze_intensity(self, data='cropped', ph_var='ph0acme', normalize=True, ylim=None):
        """
        uses the phase data information to rephase and integrate all spectra and plot a comparison
        stores the intensity information in the measurement folder
        
        data:
            'orig': self.data_orig
            'cropped': self.data

        returns: fig, intensity_data
        """
        _data = {'orig': self.data_orig,
                'cropped': self.data}[data]

        # the correction factor for the normalization is added again.
        self.phased = [ng.proc_base.ps(_data[n, :], p0=-(self.eNMRraw.loc[n, ph_var]))# + self.U_0))
                  for n in range(len(_data[:, 0]))]

        intensity = np.array([self.phased[n].real.sum() for n in range(len(_data[:, 0]))])
        
        if normalize:
            intensity /= intensity[0]

        u = [self.eNMRraw.loc[i, self._x_axis] for i, n in enumerate(self.eNMRraw[self._x_axis])]

        intensity_data = pd.DataFrame()
        intensity_data["U"] = u
        intensity_data['intensity'] = intensity
        intensity_data['ph'] = self.eNMRraw[ph_var]# + self.U_0

        self.intensity_data = intensity_data

        fig = plt.figure(figsize=(8, 6))
        _ax = plt.subplot(221)
        _ax.scatter(intensity_data['U'], intensity_data['intensity'], c='k')
        _ax.set_ylabel('intensity / a.u.')
        if normalize and (ylim is None):
            _ax.set_ylim(0,1.05)
        elif normalize:
            _ax.set_ylim(*ylim)
        _bx = plt.subplot(222, sharey=_ax)
        _bx.plot(intensity_data['intensity'], 'ok')

        _cx = plt.subplot(223, sharex=_ax)
        _cx.scatter(intensity_data['U'], intensity_data['ph'], c='k')
        
        if self.dependency.upper() == "U":
            _cx.set_xlabel('$U$ / V')
        elif self.dependency.upper() == "G":
            _cx.set_xlabel("$g$ / $($T$\cdot$m$^{-1})$")
        elif self.dependency.upper() == 'I':
            _cx.set_xlabel("$I$ / mA")
        elif self.dependency.upper() == 'RI':
            _cx.set_xlabel("$(R \cdot I)$ / V")

        _cx.set_ylabel('$\t{\Delta}\phi$ / °')

        _dx = plt.subplot(224, sharex=_bx)
        _dx.plot(intensity_data['ph'], 'ok')
        _dx.set_xlabel('vc')

        fig.savefig(self.path+'intensity_plot_'+self.expno+".pdf")

        intensity_data.to_csv(self.path+'intensity_data_'+self.expno+".csv")

        return fig, intensity_data

    def linreg(self, ulim=None, y_column='ph0'):
        """
        standard linear regression method based on the least-square method

        ulim: 
            tuple defining the voltage limits for the regression e.g. ulim = (-100, 100)
        y_column:
            column(keyword) to be analyzed from the eNMRraw dataset

        stores results in lin_res_dic[y_column]
        :returns: nothing
        """

        # select x-axis
        #self._x_axis = {"U": "U / [V]",
                #"G": "g in T/m"
                #}[self.dependency.upper()]

        # convert data
        _eNMRreg = self.eNMRraw[[self._x_axis, y_column]].sort_values(self._x_axis)

        # setting the axis for regression
        if ulim is None:
            umin = min(self.eNMRraw[self._x_axis])
        else:
            umin = ulim[0]

        if ulim is None:
            umax = max(self.eNMRraw[self._x_axis])
        else:
            umax = ulim[1]

        _nparray = np.array(_eNMRreg[(self.eNMRraw[self._x_axis] <= umax)
                                                == (self.eNMRraw[self._x_axis] >= umin)])
        _X_train, _Y_train = _nparray[:, 0], _nparray[:, 1]


        def lineareq(x, m, b):
            return x*m+b
        

        # regression object
        linmodel = lf.Model(lineareq)
        linparams = linmodel.make_params()
        linparams['b'].set(0)
        linparams['m'].set(1)
        result = linmodel.fit(_Y_train, x=_X_train, params=linparams)

        # linear parameters
        m = result.best_values['m']  # slope
        b = result.best_values['b'] # y(0)
        _Y_pred = result.best_fit
    
        # calculation of the slope deviation
        _sig_m_a = np.sqrt(np.sum((_Y_train-_Y_pred)**2)/(np.size(_Y_train)-2))
        _sig_m_b = np.sqrt(np.sum((_X_train-_X_train.mean())**2))
        sig_m = _sig_m_a/_sig_m_b

        # debug
        #print(self.sig_m)

        # R^2    
        r_square = 1 - result.residual.var() / np.var(_Y_train)
        
        self.lin_res_dic[y_column] = {'b': b,
                                'm': m,
                                'sig_m': sig_m,
                                'r_square': r_square,
                                'x': np.array(_X_train.tolist()).ravel(),
                                'y': np.array(_Y_train.tolist()).ravel(),
                                'y_fitted': _Y_pred.ravel(),
                                }
        return

    # should be replaced by a more recent function since it will be deprecated
    def lin_huber(self, epsilon=3, ulim=None, y_column='ph0'):
        """
        robust linear regression method from scikit-learn module based on the least-square method with an additional threshhold (epsilon) for outlying datapoints
        outlying datapoints are marked as red datapoints
        
        epsilon:
            threshhold > 1
        ulim: 
            tuple defining the voltage limits for the regression e.g. ulim = (-100, 100)
        y_column:
            column(keyword) to be analyzed from the eNMRraw dataset
        
        stores results in lin_res_dic[y_column]
        :returns: nothing
        """

        # select x-axis
        #self._x_axis = {"U": "U / [V]",
                   #"G": "g in T/m"
                   #}[self.dependency.upper()]

        # convert data
        _eNMRreg = self.eNMRraw[[self._x_axis, y_column]].sort_values(self._x_axis)
        
        # setting the axis for regression
        if ulim is None:
            umin = min(self.eNMRraw[self._x_axis])
        else:
            umin = ulim[0]
            
        if ulim is None:
            umax = max(self.eNMRraw[self._x_axis])
        else:
            umax = ulim[1]

        _npMatrix = np.matrix(_eNMRreg[(self.eNMRraw[self._x_axis] <= umax)
                                                 == (self.eNMRraw[self._x_axis] >= umin)])

        _X_train, _Y_train = _npMatrix[:, 0], _npMatrix[:, 1]
        
        # regression object
        huber = hub.HuberRegressor(epsilon=epsilon)
        huber.fit(_X_train, _Y_train)
        
        # linear parameters
        m = huber.coef_  # slope
        b = huber.intercept_  # y(0)
        _y_pred = huber.predict(_X_train)
        _y_pred = _y_pred.reshape(np.size(_X_train), 1)
        
        # drop the outliers
        _outX_train = np.array(_X_train[[n == False for n in huber.outliers_]])
        _outY_train = np.array(_Y_train[[n == False for n in huber.outliers_]])
        _outY_pred = np.array(_y_pred[[n == False for n in huber.outliers_]])
        
        # mark outliers in dataset
        # self._inliers = [n is not True for n in self.huber.outliers_]

        self.eNMRraw["outlier"] = True

        for n in range(len(_npMatrix[:, 0])):
            self.eNMRraw.loc[self.eNMRraw[self._x_axis] == _npMatrix[n, 0], "outlier"] = huber.outliers_[n]

        # calculation of the slope deviation
        _sig_m_a = np.sqrt(np.sum((_outY_train-_outY_pred)**2)/(np.size(_outY_train)-2))
        _sig_m_b = np.sqrt(np.sum((_outX_train-_outX_train.mean())**2))
        sig_m = _sig_m_a/_sig_m_b

        # debug
        #print(self.sig_m)

        # R^2
        r_square = huber.score(_outX_train, _outY_train)
        
        self.lin_res_dic[y_column] = {'b': b,
                                   'm': m,
                                   'r_square': r_square,
                                   'x': np.array(_X_train.tolist()).ravel(),
                                   'y': _Y_train,
                                   'y_fitted': _y_pred.ravel(),
                                   'sig_m': sig_m}

    def lin_display(self, ylim=None, show_slope_deviation=True, n_sigma_displayed=1, dpi=500, y_column='ph0', textpos=(0.5,0.15), extra_note=''):
        """
        displays the linear huber regression
        If there is an alias available from measurement object, it will replace the path in the title
        
        ylim:
            set limits of the y-axis
        
        show_slope_deviation:
            display the standard deviation
        n_sigma_displayed:
            multiplicator of the displayed standard deviation
        y_column:
            column(keyword) to be analyzed from the eNMRraw dataset
        textpos:
            tuple for the textposition
        extra_note:
            added to the text
        dpi:
            adjusts the output dpi to the required value
        
        :returns: figure
        """
        
        textx, texty = textpos
        #_x_axis = {"U":"U / [V]", "G":"g in T/m"}
        #self._x_axis = {"U":"U / [V]", "G":"g in T/m"}[self.dependency]

        print("formula: y = {0}x + {1}".format(self.lin_res_dic[y_column]['m']
                                                ,self.lin_res_dic[y_column]['b']))
        
        # create figure
        fig_enmr = plt.figure()

        # sublot phase data
        _ax = fig_enmr.add_subplot(111)
        
        # color format for outliers
        try:
            colors = ["r" if n else "k" for n in self.eNMRraw.sort_values(self._x_axis)['outlier']]
        except KeyError:
            colors = ["k" for n in self.eNMRraw[y_column]]
            print('no outliers registered')
            pass
        
        #_ax.scatter(x=np.ravel(self._eNMRreg[self._x_axis]),
                    #y=np.ravel(self._eNMRreg[y_column]),
        _ax.scatter(x=np.ravel(self.eNMRraw[self._x_axis]),
                    y=np.ravel(self.eNMRraw[y_column]),
                    marker="o",
                    c=colors)
        _ax.set_ylim(ylim)

        # format the data for plotting
        _xdata = np.ravel(self.lin_res_dic[y_column]['x'])
        _ydata = np.ravel(self.lin_res_dic[y_column]['y_fitted'])
        
        # Plot the regression
        _ax.plot(_xdata, _ydata, "r-")

        if show_slope_deviation:
            #_ax.fill_between(_xdata, _xdata*(self.m+n_sigma_displayed*self.sig_m)+self.b,
                                     #_xdata*(self.m-n_sigma_displayed*self.sig_m)+self.b,
                                     #alpha=0.5,
                                     #facecolor="blue")
            _ax.fill_between(_xdata, _xdata*(self.lin_res_dic[y_column]['m']+n_sigma_displayed*self.lin_res_dic[y_column]['sig_m'])+self.lin_res_dic[y_column]['b'],
                                     _xdata*(self.lin_res_dic[y_column]['m']-n_sigma_displayed*self.lin_res_dic[y_column]['sig_m'])+self.lin_res_dic[y_column]['b'],
                                     alpha=0.5,
                                     facecolor="blue")

        # make title
        if self.alias is None:
            title_printed = r'%s'%((self.path+self.expno).split("/")[-2]+", EXPNO: "+self.expno+extra_note)
#             plt.title('LiTFSIDOLDME') # debugging
        else:
            title_printed = self.alias+", EXPNO: "+self.expno+extra_note
#             plt.title(self.alias+", EXPNO: "+self.expno+extra_note)
#             plt.title('test2') # for debugging purposes
        plt.title(title_printed.replace('_',r' '))
    
        if self.dependency.upper() == "U":
            plt.xlabel("$U$ / V")
        elif self.dependency.upper() == "G":
            plt.xlabel("$g$ / $($T$\cdot$m$^{-1})$")
        elif self.dependency.upper() == 'I':
            plt.xlabel("$I$ / mA")
        elif self.dependency.upper() == 'RI':
            plt.xlabel("$(R \cdot I)$ / V")

        plt.ylabel("$\Delta\phi$ / °")
        
        # plotting the Textbox
        plt.text(textx, texty,
                 "y = %.4f $\cdot$ x + %4.2f\n$R^2$=%4.3f; $\sigma_m=$%4.4f"%(self.lin_res_dic[y_column]['m'],
                                                                              self.lin_res_dic[y_column]['b'],
                                                                              self.lin_res_dic[y_column]['r_square'],
                                                                              self.lin_res_dic[y_column]['sig_m']),
                 fontsize=14,
                 bbox={'facecolor':'white', 'alpha':0.7,'pad':10},
                 horizontalalignment='center',
                 verticalalignment='center',
                 transform=_ax.transAxes)

        return fig_enmr

    def lin_results_display(self, cols, regression=True, normalize=True, colors=None, markers=None, fig=None, x_legend=1.1, y_legend=1.0, ncol_legend=2, ymin=None, ymax=None, figsize=None):
        '''
        displays the phase results including the regression. Can take a previous figure from a different eNMR measurement in order to compare results
        the relegend() function can be used on these graphs to reprint the legend() for publishing
        
        cols:
            list of keywords for dataselection from the eNMRraw table
        colors:
            list of colors or a colormap --> see matplotlib documentation
        
        :returns: figure
        '''
        
        #self._x_axis = {"U":"U / [V]", "G":"g in T/m"}[self.dependency.upper()]
        
        if type(cols) != list:
            raise TypeError('cols should be a list')
        
        if fig is None:
            fig = plt.figure(figsize=figsize)
        else:
            fig = fig

        ax = fig.add_subplot(111)

        if colors is None:
            prop_cycle = plt.rcParams['axes.prop_cycle']
            colors = prop_cycle.by_key()['color']
        else:
            colors = colors
        
        if markers is None:
            markers = ['o', '^', 's', '+', '*', 'D', 'v', 'x', '<']
        else:
            markers = markers

        message = {
            'ph0acme': 'Autophase with entropy minimization',
            'ph0diff': 'Autophase with linear difference minimization',
            'ph0sqdiff': 'Autophase with square difference minimization'
            }
        
        if type(cols) == list:
            for i, col in enumerate(cols):
                if normalize:
                    corr = self.eNMRraw.loc[self.eNMRraw['U / [V]']==0, col].iloc[0]
                else:
                    corr = 0
                    
                try:
                    ax.errorbar(self.eNMRraw[self._x_axis], self.eNMRraw[col]-corr, yerr=self.eNMRraw['%s_err'%col], fmt='o', label=message[col], c=colors[i], marker=markers[i])
                except KeyError:
                    if col == 'ph0acme':
                        ax.errorbar(self.eNMRraw[self._x_axis], self.eNMRraw[col]-corr, fmt='o', label=message[col], c=colors[i], marker=markers[i])
                    else:
                        ax.errorbar(self.eNMRraw[self._x_axis], self.eNMRraw[col]-corr, yerr=self.eNMRraw['%s_err'%col], fmt='o', label='fitted data %s'%col, c=colors[i], marker=markers[i])
                except ValueError:
                    ax.errorbar(self.eNMRraw[self._x_axis], self.eNMRraw[col]-corr, fmt='o', label=message[col], c=colors[i], marker=markers[i])
                if regression:
                    ax.plot(self.lin_res_dic[col]['x'], self.lin_res_dic[col]['y_fitted']-corr, '--', label='%s lin regression'%col, c=colors[i], marker=None)

        else:
            if normalize:
                corr = self.eNMRraw.loc[self.eNMRraw['U / [V]']==0, cols][0]
            else:
                corr = 0
                    
            try:
                ax.errorbar(self.eNMRraw[self._x_axis], self.eNMRraw[cols]-corr, yerr=self.eNMRraw['%s_err'%cols], fmt='o', label='Phasefitting of Peak %s' %cols, c=colors[0], marker=markers)
                if regression:
                    ax.plot(self.lin_res_dic[cols]['x'], self.lin_res_dic[cols]['y']-corr, '--', label='%s lin regression'%cols, c=colors[0], marker=markers)
            except KeyError:
                if cols=='ph0acme':
                    ax.errorbar(self.eNMRraw[self._x_axis], self.eNMRraw[cols]-corr, fmt='o', label=message[cols], c=colors[0], marker=markers)
                else:
                    ax.errorbar(self.eNMRraw[self._x_axis], self.eNMRraw[col]-corr, yerr=self.eNMRraw['%s_err'%col], fmt='o', label='fitted data %s'%cols, c=colors[i], marker=markers)
            except ValueError:
                ax.errorbar(self.eNMRraw[self._x_axis], self.eNMRraw[cols]-corr, fmt='o', label=message[cols], c=colors[0], marker=markers)
                if regression:
                    ax.plot(self.lin_res_dic[cols]['x'], self.lin_res_dic[cols]['y_fitted']-corr, '--', label='%s lin regression'%cols, c=colors[0], marker=None)

        ax.legend(bbox_to_anchor=(x_legend, y_legend), ncol=ncol_legend)
        
        xlabel = {'U': r'$U$ / V',
                  'G': r'$g$ / (T\,m$^{-1}$)',
                  'I': '$I$ / mA',
                  'RI': '$(R \cdot I)$ / V'
                      }[self.dependency.upper()]
        
        ax.set_xlabel(xlabel)
        if normalize:
            ax.set_ylabel(r'$\phi-\phi_0$ / °')
        else:
            ax.set_ylabel(r'$\phi$ / °')
        return fig

    def plot_phcorrected_spectra(self, phasekey='ph0acme', xlim=None, save=False, savepath=None, show=True, ppm=True, orig_data=True, x_legend=1.1, y_legend= 1.0, ncol_legend=2):
        """
        plots phase corrected rows stacking the spectra to visualize the analysis result

        ppm: When True, plots the x-axis with ppm scale, xmin and xmax then are percentages
            if False, plots only the number of datapoints, xmin and xmax then are absolute values

        save: saves the spectrum in the data folder

        show: displays the spectrum {plt.show()} or not
        
        returns:
            phased data within given range
        """
        if savepath is None:
            path = self.path+"layered_spectra_"+self.expno+".png"
        else:
            path = savepath
        if orig_data:
            phasedspc = [ng.proc_base.ps(self.data_orig[n, :], p0=-(self.eNMRraw[phasekey][n]))# obsolet: +self.U_0))  # the correction factor for the normalization is added again.
                          for n in range(len(self.data_orig[:, 0]))]  # p1=self.opt[1])
        else:
            phasedspc = [ng.proc_base.ps(self.data[n, :], p0=-(self.eNMRraw[phasekey][n]))# obsolet: +self.U_0))  # the correction factor for the normalization is added again.
                          for n in range(len(self.data[:, 0]))]  # p1=self.opt[1])
            
        fig, ax = plt.subplots()
        
        if not ppm:
            for n in range(len(self.data[:, 0])):
                ax.plot(phasedspc[n].real)
            ax.set_xlim(xlim)
            ax.set_xlabel("data points")
            ax.set_ylabel("intensity (a.u.)")

        else:
            if orig_data:
                for n in range(len(self.data_orig[:, 0])):
                    ax.plot(self.ppm, phasedspc[n].real, label="row %i" %n)
            else:
                for n in range(len(self.data[:, 0])):
                    ax.plot(self.ppm, phasedspc[n].real, label="row %i" %n)
            ax.set_xlim(xlim)
            ax.set_xlabel("$\delta$ / ppm")
            ax.set_ylabel("intensity / a.u.")
            ax.legend(bbox_to_anchor=(x_legend, y_legend), ncol=ncol_legend)

        ax.ticklabel_format(style='sci', scilimits=(0,0), axis='y')

        if save:
            fig.savefig(path, dpi=500)
        if show:
            plt.show()
        phasedspc = np.asarray(phasedspc)
        
        return fig, ax
        
    def output_phase_data(self, path=None):
        """
        saves the raw phase data in the measurement-folder as a csv
        
        if path is None: saves in self.path+"phase_data_"+self.expno+".csv"
        """
        if path != None:
            self.eNMRraw.to_csv(path, sep=" ")
            return
        else:
            self.eNMRraw.to_csv(self.path+"phase_data_"+self.expno+".csv", sep=" ")
            return

    def output_results(self, path=None):
        """
        saves the mobility result data in the measurement-folder similar to obj.output_data()
        """
        from warnings import warn
        warn('this function was renamed output_properties_csv')
        
        results_output = pd.Series([self.nuc,
                                    self.mu[0],
                                    self.sig_m*self.mu[0],
                                    self.d,
                                    self.g,
                                    self.delta,
                                    self.Delta,
                                    self.uInk],
                                   index=["nucleus",
                                          "µ",
                                          "sig_µ",
                                          "d / m",
                                          "g / (T/m)",
                                          "delta / s",
                                          "Delta / s",
                                          "Uink / V"],
                                   name=self.dateipfad)
        if path is None:
            results_output.to_csv(self.path+"mobility_data_"+self.expno+".csv")
        elif path is not None:
            results_output.to_csv(path+"mobility_data_"+self.expno+".csv")
        else:
            print('ooops!')
            
    def output_properties_csv(self, path=None):
        """
        saves the mobility result data in the measurement-folder similar to obj.output_phase_data()
        """
        self.output_results(self, path)
    
    def output_all_results(self, path=None, data=False):
        """
        saves the mobility result data in the measurement-folder similar to obj.output_data()
        """
        
        from pandas import ExcelWriter
        from openpyxl import load_workbook
        
        try:
            book = load_workbook(path)
            writer = pd.ExcelWriter(path, engine='openpyxl') 
            writer.book = book
            writer.sheets = dict((ws.title, ws) for ws in book.worksheets)
        except:
            writer = ExcelWriter(path, engine='xlsxwriter')

        for key in self.lin_res_dic.keys():
            #self.mobility(key)
            results_output = pd.Series([self.nuc,
                                        self.expno,
                                        self.lin_res_dic[key]['mu'][0],
                                        self.lin_res_dic[key]['mu_err'][0],
                                        self.d,
                                        self.g,
                                        self.delta,
                                        self.Delta,
                                        self.uInk],
                                    index=["nucleus",
                                           "expno",
                                            "µ",
                                            "sig_µ",
                                            "d / m",
                                            "g / (T/m)",
                                            "delta / s",
                                            "Delta / s",
                                            "Uink / V"],
                                    name=self.dateipfad)
            results_output.to_excel(writer, sheet_name=self.nuc+'expno'+str(self.expno)+'_'+key)
        
        if data:
            try:
                self.eNMRraw.drop(['data', 'fid'], 1).to_excel(writer, sheet_name=self.nuc+'expno'+str(self.expno)+'_data')
            except KeyError:
                self.eNMRraw.to_excel(writer, sheet_name=self.nuc+'expno'+str(self.expno)+'_data')
            except:
                print('oops, an unexpected error occured')
            writer.save()
        return
    
    def mobility(self, y_column='ph0acme', electrode_distance=None, verbose=True):
        """
        calculates and returns (mobility, deviation) from the regression data
        """
        
        #self.lin_res_dic = {y_column: {'b': self.b,
                            #'m': self.m,
                            #'r^2': self.r_square,
                            #'y_reg': self._y_pred,
                            #'sig_m': self.sig_m}}
        if electrode_distance is None:
            d = self.d
        else:
            d = electrode_distance
        
        if self.dependency.upper() == "G":
            g = self.uInk
        elif self.dependency.upper() == "U":
            g = self.g
        elif self.dependency.upper() == "I":
            g = self.g
        elif self.dependency.upper() == "RI":
            g = self.g
        else:
            print("no dependency was set")
        
        if y_column is None:
            self.mu = (self.m*d)/(self.gamma*self.delta*self.Delta*g)
            #return self.mu, self.mu*(self.sig_m/self.m)

        else:
            m = self.lin_res_dic[y_column]['m']
            sig_m = self.lin_res_dic[y_column]['sig_m']
            mu = (m*d)/(self.gamma*self.delta*self.Delta*g)
            self.lin_res_dic[y_column]['mu']= mu
            self.lin_res_dic[y_column]['mu_err']= mu*(sig_m/m)
            #return self.mu, self.mu*(sig_m/m)
            #self.sig_m, self.m = sig_m, m
        
        if verbose:
            print ('%.2E (m^2/Vs)'%self.lin_res_dic[y_column]['mu'],'+- %.2E'%(self.lin_res_dic[y_column]['mu']*(self.lin_res_dic[y_column]['sig_m']/self.lin_res_dic[y_column]['m'])))
        
        return self.lin_res_dic[y_column]['mu'], self.lin_res_dic[y_column]['mu']*(self.lin_res_dic[y_column]['sig_m']/self.lin_res_dic[y_column]['m'])

